import asyncio
import os
import sys
import time
from multiprocessing import Process

from aiohttp import ClientSession
from avvyprotocol.l3_pb2 import MsgResponse, Event

from client.connect import connect_directly
from client.generators import TimedTxnManager, gen_msg_body
from core.utils import init_log, Tryings
from core.verify import SigningService

log = init_log('client')

base_dir = './load_test'


async def ws_subscription(url, ttm: TimedTxnManager):
    async with ClientSession().ws_connect(url) as ws:
        log.info(f'WS {url} connected!')
        msgs = 0
        async for msg in ws:
            msgs += 1
            if msgs % 100 == 0:
                log.info(f'from {url} received {msgs} subscribed msg')
            if msg.data:
                try:
                    e = Event()
                    e.ParseFromString(msg.data)

                    if e.msg.body:
                        ttm.register_subscribed_msg(e.msg)
                    else:
                        resp = MsgResponse()
                        resp.ParseFromString(msg.data)

                        ttm.register_subscribed_msg(resp)

                except Exception as e:
                    log.exception(e)


async def process_ws_subscriptions(urls, ttm: TimedTxnManager):
    tsks = []
    for url in urls:
        tsks.append(asyncio.create_task(ws_subscription(url, ttm)))

    for tsk in tsks:
        await tsk


async def communicate(uri, ttm, tx_count, msg_count, data):
    async with connect_directly(uri) as stub:
        for txn_id, txn in enumerate(ttm.generate_txs(tx_count, msg_count, data)):
            resp_tx = await stub.ProcessTxn(txn)
            if resp_tx.body.responses[0].status == MsgResponse.Status.REDIRECT:
                return txn_id, resp_tx.body.responses[0].body['redirect_uri'].decode()
            ttm.register_response(resp_tx)


async def grpc_communication(uri, ttm, tx_count, msg_count, data):
    _processed_txns = 0
    _tx_count = tx_count
    _uri = uri
    while 1:
        _tx_count -= _processed_txns

        result = await communicate(_uri, ttm, _tx_count, msg_count, data)
        if result:
            txn_id, redirect_uri = result
            log.info(f'Have redirect at: {txn_id} to {redirect_uri}')
            _processed_txns += txn_id
            _uri = redirect_uri
        else:
            break

    return True


def process_client(grpc_uri, ws_uri, sign_keeper):
    log.info(f'Start client to: {grpc_uri}')
    ttm = TimedTxnManager(signing_service=sign_keeper)

    loop = asyncio.get_event_loop()
    ws_subs = loop.create_task(process_ws_subscriptions(ws_uri, ttm))

    def run_process_tx(tx_count, msg_count):
        return loop.run_until_complete(
            asyncio.ensure_future(
                grpc_communication(grpc_uri, ttm, tx_count, msg_count, gen_msg_body)
            )
        )

    loop.run_until_complete(
        asyncio.sleep(3)
    )

    for tx_count, msg_count in [(1, 1)]:
        _start = time.time()
        with Tryings(10, run_process_tx, tx_count, msg_count, ) as res:
            if res:
                log.info(f'Sent {tx_count, msg_count} at {time.time() - _start}')
            else:
                log.info(f'({tx_count, msg_count}) processing failed :(')

    while not ttm.is_all_txns_finished or not ttm.is_all_objects_finished:
        time.sleep(3)
        wait_times = {k: time.time() - v["started"] for k, v in ttm.not_finished_objects().items() if 'started' in v}
        log.info(f'Wait \n{wait_times}')
        if not wait_times:
            break

    finished_times = {k: v["dt"] for k, v in ttm.finished_objects().items()}
    log.info(f'Finished: \n{finished_times}')
    ws_subs.cancel()
    ttm.dump(base_dir)


def start_client(sign_keeper):
    grpc_uri = f'localhost:5051'
    ws_uris = [
        f'http://node1:54621/ws/{sign_keeper.pub.decode()}',
        f'http://node2:54622/ws/{sign_keeper.pub.decode()}',
        f'http://node3:54623/ws/{sign_keeper.pub.decode()}',
        f'http://node4:54624/ws/{sign_keeper.pub.decode()}',
    ]
    process_client(grpc_uri, ws_uris, sign_keeper)


def start_load_client(number):
    os.makedirs(base_dir, exist_ok=True)
    signing_service = SigningService(f'{base_dir}/client_{number}')
    start_client(signing_service)


if __name__ == '__main__':
    count = int(sys.argv[1]) if len(sys.argv) == 2 else 1
    procs = []
    for i in range(count):
        p = Process(target=start_load_client, args=(i,))
        p.daemon = True
        p.start()
        procs.append(p)
        log.info(f'Client {i} started')

    for i, p in enumerate(procs):
        p.join()
        log.info(f'Client {i} finished')
