import json
import random
import time
from collections import defaultdict

import numpy as np
from avvyprotocol.l3_pb2 import Msg, TxnRequestBody, TxnRequest, TxnResponseBody

from core.subscribe.filters import SubscribeFiltersEncoder
from core.utils import generate_id, init_log, dumps
from core.verify import SigningService

np_array = lambda x: np.ones(16) * x
np_matrix = lambda x: np.ones((4, 4)) * x


def gen_subscribe_filter(f, _min=None, _max=None, _eq=None, _more=None, _less=None):
    predicates = []
    if _min and _max:
        predicates.append({'between': [f(_min), f(_max)]})
    if _eq:
        predicates.append({'=': f(_eq)})
    if _more:
        predicates.append({'>': f(_more)})
    if _less:
        predicates.append({'<': f(_less)})

    if len(predicates) > 0 and len(predicates) % 2 == 0:
        return [{'_': predicates[:len(predicates) / 2]}, {'_': predicates[len(predicates) / 2:]}]

    return [{'_': predicates}]


def gen_chld(parent):
    return {'object_id': generate_id(),
            'object_class': 0,
            'transform': np_matrix(random.random()),
            'parent': parent}


def gen_msg_body(size):
    object_id = random.choice([generate_id(), 1, -1, 0, None, generate_id(), 100500, 100600, 100700])
    return {
        'object_id': object_id,
        'object_class': random.choice([1, 0, -1, 100500, ]),
        'transform': np_matrix(4) * random.random(),
        'view_id': random.randint(-10, 10),
        'type': random.choice(['block', 'struct', 'tree', 'test']),
        'parent': random.choice([1, -1, 0, None, generate_id(), 100500]),
        'children': [gen_chld(object_id) for _ in range(size)]
    }


default_subscribe_filters = [
    {'object_id': [{'>': 0}]},
    {'view_id': [{'between': [5, 5]}, {'=': 9}]},
    {'transform': [{'between': [np_array(50), np_array(70)]}, ]},
    {'parent': [{'=': 100500}, {'=': -1}]},
]


def make_subscribe_msg(s_filter_data):
    return make_msg({'subscribe_filters': json.dumps(s_filter_data, cls=SubscribeFiltersEncoder).encode()},
                    type='subscribe')


def make_add_object_msg(data: dict):
    return make_msg(data)


def make_msg(data: dict, type=None):
    body = {k: dumps(v) for k, v in data.items()}
    body['type'] = (type or 'add_object').encode()
    return Msg(
        id=generate_id(),
        body=body
    )


def make_txn_request(msgs, signing_service: SigningService):
    body = TxnRequestBody(
        txn_id=generate_id(),
        requests=msgs
    )
    return TxnRequest(
        body=body,
        auth_info=signing_service.get_auth_info(body)
    )


log = init_log('client')


class TimedTxnManager:
    def __init__(self, signing_service: SigningService):
        self.signing_service = signing_service
        self.sent_txns = defaultdict(dict)
        self.sent_objects = defaultdict(dict)

    def generate_txs(self, tx_count=1, msg_count=1, data=None, msg_factory=make_add_object_msg, msg_size=10):
        for i in range(tx_count):
            tx_msgs = {}
            for j in range(msg_count):
                _data = (data if not callable(data) else data(msg_size)) or {}
                msg = msg_factory(_data)
                tx_msgs[msg.id] = msg

                if msg_factory == make_add_object_msg and 'object_id' in _data:
                    self.sent_objects[_data['object_id']] = {'size': len(msg.SerializeToString()), 'start': time.time()}

            tx = make_txn_request(list(tx_msgs.values()), self.signing_service)
            self.sent_txns[tx.body.txn_id] = {'size': len(tx.body.SerializeToString()), 'start': time.time(), }
            log.info(f'Tx generated: {tx.body.txn_id}')
            yield tx

    def register_response(self, response: TxnResponseBody):
        send_info = self.sent_txns.get(response.body.txn_id)
        if not send_info: log.warn(f'Received not sent txn_id {response.body.txn_id}')

        self.sent_txns[response.body.txn_id]['dt'] = time.time() - send_info["start"]

    def register_subscribed_msg(self, msg):
        object_id = msg.body.get('object_id')
        if not object_id:
            return
        try:
            object_id = int(object_id.decode())
        except Exception as e:
            return
        data = self.sent_objects.get(object_id)
        if data:
            dt = time.time() - data["start"]
            self.sent_objects[object_id]['dt'] = dt

    def dump(self, base_dir):
        with open(f'{base_dir}/transactions_times.csv', 'a') as f:
            for txn_id, data in self.sent_txns.items():
                if 'dt' in data:
                    f.write(f'{self.signing_service.pub.decode()};{data["size"]};{data["dt"]};\n')
                else:
                    log.warn(f'txn: {txn_id} has not respond')

        with open(f'{base_dir}/objects_times.csv', 'a') as f:
            for oid, data in self.sent_objects.items():
                if 'dt' in data:
                    f.write(f'{self.signing_service.pub.decode()};{data["size"]};{data["dt"]};\n')
                else:
                    log.warn(f'obj: {oid}  {data} has not respond')

    @property
    def is_all_objects_finished(self):
        for oid, info in self.sent_objects.items():
            if 'dt' not in info:
                return False
        return True

    @property
    def is_all_txns_finished(self):
        for txn_id, info in self.sent_txns.items():
            if 'dt' not in info:
                return False
        return True

    def not_finished_objects(self):
        return {oid: info for oid, info in self.sent_objects.items() if 'dt' not in info}

    def finished_objects(self):
        return {oid: info for oid, info in self.sent_objects.items() if 'dt' in info}
