from functools import partial
from threading import Thread
from time import sleep
from typing import Tuple, List

import yaml
from avvyprotocol.l3_pb2 import TxnRequest
from pysyncobj import SyncObj, SyncObjConf, replicated

import config as config
from config import CONSENSUS_NODES_FILE, CURRENT_NODE, NODE_INFO
from core.transactions_processing import TxnsProcessor
from core.utils import init_log
from load_test.utils import set_timestamp

log = init_log()


class NodesConfig:
    """
    Для предоставления информации по узлам рафт кластера.
    """

    def __init__(self, cfg_file=None, current_node=None):
        if not cfg_file:
            cfg_file = open(CONSENSUS_NODES_FILE)

        with cfg_file as f:
            nodes = yaml.load(f.read(), Loader=yaml.FullLoader)['nodes']

        self.another_nodes = {}
        for node in nodes:
            self.another_nodes[node['host']] = node
        node_info = NODE_INFO(current_node or CURRENT_NODE)
        self.current_host, self.current_raft_port, self.current_grpc_port = node_info['host'], node_info['raft_port'], \
                                                                            node_info['grpc_port']

        if self.current_host in self.another_nodes:
            self.another_nodes.pop(self.current_host)

    def get_raft_connection_info(self) -> Tuple[str, List[str]]:
        return f'{self.current_host}:{self.current_raft_port}', \
               [f'{host}:{node["raft_port"]}' for host, node in self.another_nodes.items()]

    def get_grpc_uri(self, raft_uri):
        m_host, m_r_port = raft_uri.split(':')
        master_node = self.another_nodes.get(m_host)
        return f'{master_node["host"]}:{master_node["grpc_port"]}'


nodes_cfg = NodesConfig()


class Consensus(SyncObj):
    """
    Синхронизация транзакций между узлами рафт кластера и запуск последующей обработки..
    NB! Метод с @replicated гарантировано запустится у всех нод с одинаковыми параметрами, но может выполнятся
    на каждом узле за разное время.

    :param txns_proc - процессор транзакций которому будут передаваться синхронизированные транзакции
    :param process_txn_cb - то что выполнится на стадии синхронизации
    :param replicate_txn_cb - то что выполнится после стадии синхронизации и лишь на запустившей ноде

    """

    def __init__(self, txns_proc: TxnsProcessor, current_node=None, other_nodes=None):
        if current_node is None and other_nodes is None:
            c_node, o_nodes = nodes_cfg.get_raft_connection_info()
        else:
            c_node, o_nodes = current_node, other_nodes

        self.__transactions = []

        self.__txns_proc = txns_proc
        self.__sync_proc_debt = 0

        super().__init__(selfNode=c_node,
                         otherNodes=o_nodes,
                         conf=SyncObjConf(
                             password=config.CONSENSUS_PASSWORD,
                             # dynamicMembershipChange=True,
                             sendBufferSize=config.SYNC_BUFFER_SIZE,
                             recvBufferSize=config.SYNC_BUFFER_SIZE,
                             appendEntriesBatchSizeBytes=config.SYNC_BUFFER_SIZE,
                             commandsQueueSize=config.SYNC_BUFFER_SIZE,
                             onStateChanged=lambda o, n: log.info(f'State changed: {o} -> {n}\n{self.getStatus()}'),
                             onReady=lambda: log.info(f'Ready!\n{self.getStatus()}')
                         ))

        log.info(f'Start init consensus...')
        try:
            def wait():
                self.waitBinded()
                self.waitReady()

                while self.getStatus()['leader'] is None:
                    log.info(f'Wait leader...\n{self.getStatus()}')
                    sleep(1)
                log.info(f'FINAL STATE:{self.getStatus()}')

            th = Thread(target=wait)
            th.setDaemon(True)
            th.start()

        except Exception as e:
            log.exception(e)
            raise e

    def __sync_cb(self, txn, res, err):
        if err:
            log.error(f'Error proc tx: {txn.body.tx_nonce}')

        self.__sync_proc_debt -= 1

    def sync_transaction(self, txn: TxnRequest):
        set_timestamp(txn, '_tst_consensus_start')
        self._add_transaction(txn, callback=partial(self.__sync_cb, txn))
        self.__sync_proc_debt += 1

    @property
    def sync_proc_debt(self):
        """
        Количество транзакций в очереди на синхронизацию.
        """
        return self.__sync_proc_debt

    @replicated
    def _add_transaction(self, txn: TxnRequest):
        """
        Реплицируемый метод добавления новой транзакции и последующей ее обработки.
        """
        self.__transactions.append(txn)
        self.__txns_proc.add_txn_to_processing(txn)
        return txn

    def get_leader_grpc_conn(self):
        leader_uri = self._getLeader()
        if leader_uri and leader_uri != self.selfNode:
            return nodes_cfg.get_grpc_uri(leader_uri.address)

    def _print_status(self):
        st = self.getStatus()
        log.info(
            f'Status of {st["self"]}: master:[{st["leader"]}], uptime:{st["uptime"]}, commit_idx:{st["commit_idx"]}')
