from typing import Tuple, List

from avvyprotocol.l3_pb2 import Msg, MsgResponse, Event, Signer

from config import IIDAP_ID
from core.store.prove import Storage
from core.subscribe.event_listeners import EventListeners
from core.subscribe.filters import SubscribeFilter
from core.utils import init_log

log = init_log()


class Handler:
    """
    Base class for handling msgs. Everyone must implement their execute method logic.
    """
    _storage: Storage = Storage(IIDAP_ID)
    msg_type: bytes = None

    def execute(self, client: Signer, msg: Msg) -> Tuple[MsgResponse, List[Event]]:
        raise NotImplementedError()


class SubscribeHandler(Handler):
    """
    Change subscribe filter for client
    """
    msg_type = b'subscribe'

    def __init__(self):
        self.event_listeners = EventListeners()

    def execute(self, client: Signer, msg: Msg) -> Tuple[MsgResponse, List[Event]]:
        self.event_listeners.set_client_subscribe_filter(client.public_key.decode(), SubscribeFilter(msg))
        return MsgResponse(status=MsgResponse.Status.ACCEPTED), []


class AddObjectHandler(Handler):
    """
    Add object to storage, and if raises - return DECLINED
    """
    msg_type = b'add_object'

    def execute(self, client: Signer, msg: Msg) -> Tuple[MsgResponse, List[Event]]:
        try:
            self._storage.add_dict(msg.body)
            return MsgResponse(status=MsgResponse.Status.ACCEPTED, body=msg.body, msg_id=msg.id), []
        except Exception as e:
            log.exception(e)
            return MsgResponse(status=MsgResponse.Status.DECLINED, body={'error': f'exception {e}'.encode()}), []


handlers = [AddObjectHandler, SubscribeHandler]
