import asyncio
from asyncio import CancelledError, sleep
from typing import List, Union

from avvyprotocol.l2_pb2 import Rollup
from avvyprotocol.l3_pb2 import TxnRequest, Msg

from config import IIDAP_ID, ROLLUP_MAX_TXN_COUNT, TICK_FRAME_RATE_S, SHUTDOWN_WAIT_TICK_COUNT
from core.store.data import merkle_root_ipfs_store
from core.store.prove import Storage
from core.subscribe.event_listeners import EventListeners
from core.utils import Tryings, generate_id, init_log
from l2.client import L2ClientInterface, receive_events
from l2.utils import parse_event_data

log = init_log('rollup')


class RollupProcessor:
    """
    Сущность отвечающая за формирование роллапов, отправку
    """

    def __init__(self, storage: Storage,
                 event_listeners: EventListeners,
                 start_engine,
                 consensus,
                 l2_client: L2ClientInterface):

        self.storage = storage
        with Tryings(4, self.storage.get_root) as res:
            self.prev_hash = res
            log.info(f'Initialise rollup. Root: {self.prev_hash}')

        self.txns_list: List[TxnRequest] = []

        self.consensus = consensus
        self.l2_client = l2_client
        self.start_engine = start_engine

        self.status_proc = RollupDigestProcessor(event_listeners=event_listeners, start_engine=start_engine)

        self._wait_counter = 0
        self._watch_dog_task = asyncio.get_event_loop().create_task(self.watch_dog())

    async def append_txn(self, txn: TxnRequest):
        """
        Учитывание транзакции в списке роллапа.
        """
        self._wait_counter = 0
        if not self.consensus._isLeader():
            return

        rlp = self._append_txn(txn)
        if rlp:
            log.info(f'Time to send rollup!\n {rlp.prev_hash} + {len(rlp.txns_list)}txs -> {rlp.current_hash} txs')
            rollup_digest_id = await self.l2_client.send_rollup_async(rlp)
            log.info(f'Rollup sent. RD_ID: {rollup_digest_id}')

    async def at_finish(self):
        """
        Немедленное формирование и отправка роллапа
        """
        if self.prev_hash != self.storage.get_root():
            rollup = self._make_rollup()
            await self.l2_client.send_rollup_async(rollup)

    def _append_txn(self, txn: TxnRequest) -> Union[Rollup, None]:
        self.txns_list.append(txn)
        if len(self.txns_list) == ROLLUP_MAX_TXN_COUNT:
            return self._make_rollup()

    def _make_rollup(self) -> Rollup:
        data_source_uri = merkle_root_ipfs_store.put_snapshot(self.storage.file_path)
        prev_hash, current_hash, txns_list = self.prev_hash, self.storage.get_root(), self.txns_list[:]
        self.prev_hash, self.txns_list = current_hash, []
        return Rollup(id=generate_id(),
                      prev_hash=prev_hash,
                      current_hash=current_hash,
                      txns_list=txns_list,
                      data_source_uri=data_source_uri)

    async def watch_dog(self):
        while 1:
            await sleep(TICK_FRAME_RATE_S)
            self._wait_counter += 1
            if self._wait_counter == SHUTDOWN_WAIT_TICK_COUNT:
                log.info(f'Time to stop.')
                break

        await self.at_finish()
        self.start_engine.event_shutdown()

    def stop_task(self):
        self._watch_dog_task.cancel()
        self.status_proc.stop_task()


class RollupDigestProcessor:
    def __init__(self, event_listeners, start_engine, loop=None):
        self.event_listeners = event_listeners
        self.start_engine = start_engine

        loop = loop or asyncio.get_event_loop()
        self._tsk = loop.create_task(self.__process_rollup_statuses())
        self._sent_rollups_digests = {}

    def stop_task(self):
        self._tsk.cancel()

    async def __process_rollup_statuses(self):
        """
        Слежение за статусами роллап дайджестов. Если роллап отвергнут, происходит инициация перезагрузки узла.
        """
        declined = False
        try:
            async for event in receive_events('rollup_digest', operator='CONTAINS'):
                log.info(f'Rollup digest changed: {event}')
                event = parse_event_data(event['data']['value']['TxResult']['result']['events'])
                if event['RD_IIDAP_ID'] != IIDAP_ID:
                    continue

                new_status = event['RD_STATUS']
                rollup_digest_id = int(event['RD_ID'])
                self._sent_rollups_digests[rollup_digest_id] = new_status

                if new_status == 'DECLINED':  # todo make normal status model
                    await self.event_listeners.broadcast_msg(Msg(id=generate_id(), body={'type': b'RESTART'}))
                    declined = True
                    break

                elif new_status == 'ACCEPTED':  # todo make normal status model
                    del self._sent_rollups_digests[rollup_digest_id]
                    continue

        except CancelledError:
            log.info(f'Stop listen')
        except Exception as e:
            log.exception(e)
        finally:
            if declined:
                self.start_engine.event_restart()
