import asyncio
import sys
from asyncio import Event, Queue

from config import IIDAP_ID, NODE_INFO, CURRENT_NODE
from core.consensus import Consensus
from core.rollup import RollupProcessor
from core.store.data import merkle_root_ipfs_store
from core.store.prove import Storage
from core.subscribe.event_listeners import EventListeners
from core.transactions_processing import TxnsProcessor
from core.utils import init_log
from grpc_server import start_grpc_server
from l2.subproc import SubProcL2Client
from ws.server import start_ws_server

log = init_log('startup')


class StartEngine:

    def __init__(self, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.node_info = NODE_INFO(CURRENT_NODE)

        self._shutdown_queue = Queue()

    def _load_last_actual_state(self) -> Storage:
        try:
            rollup_digest = self.l2_client.get_last_accepted_rollup_digest()
        except Exception as e:
            log.exception(e)
            sys.exit(-1)

        if rollup_digest:
            lres = merkle_root_ipfs_store.load_snapshot(rollup_digest['dsUri'])
            storage = Storage(IIDAP_ID)
            log.info(f'Last rollup digest: {rollup_digest}'
                     f'\nsnapshot loaded: {lres}\ncurrent root:{storage.get_root()}')
            return lres
        else:
            log.info(f'Initialise {IIDAP_ID}. No accepted rollups digests!')

    def start_iidap(self):
        """
        Если стартануто - выключаем и начинаем стартовать:
            Регистрация, загрузка последнего актуального состояния, инстанциация всех сущностей. запуск серверов

        """
        self.l2_client = SubProcL2Client()
        self.l2_client.l3_register()

        self._load_last_actual_state()

        self.storage = Storage(IIDAP_ID)
        self.event_listeners = EventListeners()
        self.txns_proc = TxnsProcessor(loop=self.loop, event_listeners=self.event_listeners)
        log.info(f'Consensus will start...')
        self.consensus_proc = Consensus(txns_proc=self.txns_proc)
        log.info(f'Rollup proc will start...')
        self.rollup_proc = RollupProcessor(storage=self.storage,
                                           event_listeners=self.event_listeners,
                                           start_engine=self,
                                           consensus=self.consensus_proc,
                                           l2_client=self.l2_client)
        self.txns_proc.set_rollup_processor(self.rollup_proc)

        log.info(f'GRPC server will start...')
        self.grpc_serv_task = self.loop.create_task(
            start_grpc_server(consensus_processor=self.consensus_proc, port=self.node_info["grpc_port"])
        )

        log.info(f'WS server will start...')
        self.ws_serv_task = self.loop.create_task(
            start_ws_server(event_listeners=self.event_listeners, port=self.node_info["ws_port"], )
        )

    async def shutdown_iidap(self):
        # log.info(f'consensus destroyed...')
        # self.grpc_serv_task.cancel()
        # log.info(f'grpc stopped...')
        #
        # self.ws_serv_task.cancel()
        # log.info(f'ws stopped...')

        self.consensus_proc.destroy()
        log.info(f'consensus destroyed...')

        await self.txns_proc.stop_process_txn_queue_task()
        log.info(f'txn proc stopped... ')

        await self.rollup_proc.stop_task()
        log.info(f'rollup stop watchdog...')

        await self.rollup_proc.status_proc.stop_task()
        log.info(f'rollup statuses processing stopped...')

        await self.shutdown_all_tasks()
        self.loop.stop()
        log.info(f'loop stopped,')
        self.loop.close()
        log.info(f'loop closed.')

    def event_restart(self):
        self._shutdown_queue.put_nowait('RESTART')

    def event_shutdown(self):
        self._shutdown_queue.put_nowait('SHUTDOWN')

    async def main_loop(self):
        while 1:
            self.start_iidap()
            event = await self._shutdown_queue.get()
            await self.shutdown_iidap()
            if event == 'RESTART':
                continue
            elif event == 'SHUTDOWN':
                break

    def start_node(self):
        self.loop.run_until_complete(self.main_loop())

    async def shutdown_all_tasks(self, ):
        tasks = [task for task in asyncio.tasks.all_tasks() if task is not
                 asyncio.tasks.current_task()]
        list(map(lambda task: task.cancel(), tasks))
        results = await asyncio.gather(*tasks, return_exceptions=True)
        print('finished awaiting cancelled tasks, results: {0}'.format(results))
