import time
from typing import List, MutableMapping

from avvyprotocol.l3_pb2 import Event

from core.subscribe.filters import SubscribeFilter
from core.utils import init_log, loads
from ws.connection import ClientConnection

log = init_log()


class EventListeners:
    """
    Прослойка между веб-сок сервером и процессингом транзакций.
    С конфигурованием фильтраций и подключений
    """
    client_connections: MutableMapping[bytes, ClientConnection] = {}
    client_filters: MutableMapping[bytes, SubscribeFilter] = {}

    async def broadcast_subscribed_msg_events(self, events: List[Event]):
        """
        Отправить события инициатору и всем сфильтровавшим остальным клиентам
        """
        for event in events:
            e_client_id = event.initiator.public_key.decode()
            if e_client_id in self.client_connections:
                await self.send_directly_responses(e_client_id, [event.msg])

            msg_data = {k: loads(v) for k, v in event.msg.body.items()}
            for c, conn in self.client_connections.items():
                if c != e_client_id and c in self.client_filters and self.client_filters[c].filter_data(msg_data):
                    await self.send_directly_responses(c, [event.msg])

    async def broadcast_msg(self, msg):
        """
        Отправить сообщение всем подключенным участникам
        """
        for c, conn in self.client_connections.items():
            await conn.send_message(msg)

    async def send_directly_responses(self, client_key, responses: List):
        """
        Отправить события только инициатору
        """
        if client_key not in self.client_connections:
            log.warning(f'Client is not present')
            return

        for response in responses:
            response.body['_tst_ws_send'] = f'{time.time()}'.encode()
            try:
                await self.client_connections[client_key].send_message(response)
            except ConnectionResetError as e:
                log.warn(f'closing connection: {client_key}, because: {e}')
                await self.close_client_connection(client_key)
                break
            except Exception as e:
                log.exception(e)

    def set_client_subscribe_filter(self, client_key, s_filter: SubscribeFilter):
        """
        Установить фильтр для клиента
        """
        self.client_filters[client_key] = s_filter

    def set_client_connection(self, client_key, conn: ClientConnection):
        """
        Добавить в учет клиентское соединение
        """
        log.info(f'add client connection {client_key}')
        self.client_connections[client_key] = conn

    async def close_client_connection(self, client_key):
        try:
            await self.client_connections[client_key].close_connection()
        except Exception as e:
            log.error(f'Error at close client: {client_key} {e}')
            log.exception(e)

        try:
            del self.client_connections[client_key]
        except Exception as e:
            log.error(f'Error ar delete client: {client_key} {e}')
            log.exception(e)

        if client_key in self.client_filters:
            del self.client_filters[client_key]

    async def close_all_connections(self):
        for cl in self.client_connections.values():
            await self.close_client_connection(cl)
