import json
from collections import defaultdict
from typing import Any

import numpy as np

from core.utils import init_log, loads
from avvyprotocol.l3_pb2 import Msg

log = init_log()
pred_names = {
    '>': lambda x, y: x > y,
    '<': lambda x, y: x < y,
    '>=': lambda x, y: x >= y,
    '<=': lambda x, y: x <= y,
    '=': lambda x, y: x == y
}


class Predicate:

    def apply(self, value, another):
        raise NotImplemented


class BinaryPredicate(Predicate):
    def __init__(self, name):
        self.pred = pred_names[name]
        self.name = name

    def apply(self, value, another):
        if type(value) == type(another):
            if isinstance(another, np.ndarray):
                if another.shape == value.shape:
                    return (self.pred(another, value)).all()
                return False
            else:
                return self.pred(another, value)
        return False


class Between(Predicate):

    def apply(self, value, another):
        if type(another) == type(value['min']) == type(value['max']):
            if isinstance(another, np.ndarray):
                if another.shape == value['min'].shape and another.shape == value['min'].shape:
                    return (value['min'] <= another).all() and (another <= value['max']).all()
                return False
            else:
                return value['min'] <= another <= value['max']
        return False


def predicate_factory(data) -> (Predicate, Any):
    k, v = list(data.items())[0]

    if k in pred_names:
        return BinaryPredicate(k), v
    else:
        assert type(v[0]) == type(v[1])
        return Between(), {'min': v[0], 'max': v[1]}


class SubscribePropFilter:
    def __init__(self, raw_data):
        """ It is array of predicates and values
        [
            {'>':'value'},
            {'between': [min, max]},
            {'=':'another_value'}
        ]
        Applied at property value with AND .
        """
        self.predicates_values = []
        for raw_pred in raw_data:
            try:
                self.predicates_values.append(
                    predicate_factory(raw_pred)
                )
            except Exception as e:
                log.error(f'can not compile filter by: {raw_pred}')
                log.exception(e)

    def apply(self, prop_value):
        return all([pred.apply(val, prop_value) for pred, val in self.predicates_values])


class SubscribeFilter:
    def __init__(self, msg: Msg):
        """
        Example of filter data:
        [
            {prop_name: filters, another_prop_name: another_filters},
            {prop_name: another_filters},

        ]
        Each filters is a data for SubscribePropFilter.
        Applied for message properties dict with OR
        """
        self.prop_filters = defaultdict(list)
        filters_data = parse_subscribe_filter(msg)
        for fdat in filters_data:
            for prop_name, raw_filters in fdat.items():
                self.prop_filters[prop_name].append(SubscribePropFilter(raw_filters))

    @property
    def disabled(self):
        return len(self.prop_filters) == 0

    def filter_data(self, msg_props):
        for p_name, p_value in msg_props.items():
            filters = self.prop_filters.get(p_name)
            if filters and any([f.apply(p_value) for f in filters]):
                return True


class SubscribeFiltersEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return {'__nd_array': obj.tolist()}
        return json.JSONEncoder.default(self, obj)


def subscribe_filter_decoder_hook(obj):
    for k, v in obj.items():
        if k == '__nd_array':
            return np.array(v)
    return obj


def parse_subscribe_filter(msg: Msg):
    filter_bytes = msg.body['subscribe_filters']
    return loads(filter_bytes)
