from asyncio import Queue, CancelledError
from typing import Dict, Tuple, List

from avvyprotocol.l3_pb2 import Signer, Msg, MsgResponse, Event, TxnRequest

from core.handlers import Handler, handlers
from core.rollup import RollupProcessor
from core.subscribe.event_listeners import EventListeners
from core.utils import init_log
from load_test.utils import set_timestamp

log = init_log('txns')


class TxnsProcessor:
    """
    Процессит хендлерами сообщения транзакций.
    При инициализации запускает фоном обработку очереди транзакций.
    """
    txn_msg_routing_rules: Dict[bytes, Handler] = {hdlr.msg_type: hdlr() for hdlr in handlers}

    def __init__(self, loop, event_listeners: EventListeners):
        self.process_txn_queue_task = loop.create_task(self.process_txn_queue())
        self.txns_queue = Queue()

        self.event_listeners = event_listeners

        self.rollup_proc: RollupProcessor

    def set_rollup_processor(self, rollup_proc: RollupProcessor):
        self.rollup_proc = rollup_proc

    async def stop_process_txn_queue_task(self):
        await self.txns_queue.join()
        self.process_txn_queue_task.cancel()

    async def process_txn(self, txn: TxnRequest):
        """
        Обработка транзакции: роутинг сообщений на хендлеры и отправка эвентов
        """
        client = txn.auth_info.signers[0]

        all_responses = []
        all_events = []
        for msg in txn.body.requests:
            resp, events = self._route_msg(client, msg)
            all_responses.append(resp)
            all_events.extend(events)

        await self.event_listeners.send_directly_responses(client.public_key.decode(), all_responses)
        await self.event_listeners.broadcast_subscribed_msg_events(all_events)

    def _route_msg(self, client: Signer, msg: Msg) -> Tuple[MsgResponse, List[Event]]:
        msg_type = msg.body.get('type')
        if not msg_type:
            return MsgResponse(status=MsgResponse.Status.DECLINED,
                               body={'error': f'Need type in msg.body'}), []
        hdlr = self.txn_msg_routing_rules.get(msg_type)
        if not hdlr:
            return MsgResponse(status=MsgResponse.Status.DECLINED,
                               body={'error': f'No handlers for message type: {msg.body.get("type")}'}), []
        return hdlr.execute(client, msg)

    async def process_txn_queue(self):
        """
        Обработка очереди транзакции
        """
        while 1:
            try:
                txn = await self.txns_queue.get()
                log.info(f'q>>> {txn.body.txn_id}')
                set_timestamp(txn, '_tst_processing_start')
            except CancelledError as e:
                break

            try:
                await self.process_txn(txn)
            except Exception as e:
                log.exception(e)

            try:
                await self.rollup_proc.append_txn(txn)
            except Exception as e:
                log.exception(e)

            self.txns_queue.task_done()

    def add_txn_to_processing(self, txn):
        """
        Добавление транзакции в очередь на обработку
        """
        log.info(f'TX proc add txn: {txn.body.txn_id} >>>q')
        self.txns_queue.put_nowait(txn)
