import json
import logging
import os
import random
import sys
from logging.handlers import TimedRotatingFileHandler
from time import sleep

import numpy as np

import config as config

logs_path = config.LOG_PATH or os.path.join(os.getcwd(), 'logs')


def init_log(name='main', level=config.LOG_LEVEL, only_stream=False, only_file=False):
    log = logging.getLogger(name)

    if len(log.handlers) == 0:
        fmt = '%(asctime)s %(levelname)s %(module)s:%(lineno)d [%(process)d]: %(message)s'
        if not only_file:
            sh = logging.StreamHandler(sys.stdout)
            sh.setFormatter(logging.Formatter(fmt))
            log.addHandler(sh)

        if not only_stream:
            os.makedirs(logs_path, exist_ok=True)

            trfh = TimedRotatingFileHandler(os.path.join(logs_path, f'{name}.log'), when='D', backupCount=7)
            trfh.setFormatter(logging.Formatter(fmt))
            log.addHandler(trfh)

        log.setLevel(level)
    return log


log = init_log()


class Tryings:
    """
    Класс для возможности запустить функцию несколько раз с увеличивающимся ожиданием
    Например:

    >>> with Tryings(2, lambda x:x*3, 1) as result: print(result)
    2

    """

    def __init__(self, try_count, f, *args, **kwargs):
        self.try_count, self.f, self.args, self.kwargs = try_count, f, args, kwargs

    def __enter__(self):
        self.tryings = 0
        while 1:
            try:
                return self.f(*self.args, **self.kwargs)
            except Exception as e:
                log.exception(e)
                sleep(self.tryings * 1.4)
                self.tryings += 1
                if self.tryings > self.try_count:
                    return

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class NdArrayEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return {'__nd_array': obj.tolist()}
        if isinstance(obj, bytes):
            return obj.decode()
        return json.JSONEncoder.default(self, obj)


def decoder_hook(obj):
    for k, v in obj.items():
        if k == '__nd_array':
            return np.array(v)

    return obj


def dumps(data):
    return json.dumps(data, cls=NdArrayEncoder).encode()


def loads(data):
    if data is not None:
        try:
            while isinstance(data, (str, bytes)):
                data = json.loads(data, object_hook=decoder_hook)
            return data
        except Exception as e:
            try:
                return float(data)
            except ValueError:
                try:
                    return int(data)
                except ValueError:
                    try:
                        return str(data)
                    except ValueError:
                        return data
    return None


def generate_id():
    return random.randint(1, 1 << 63)
