from hashlib import sha256

from fastecdsa import curve, ecdsa, keys
from fastecdsa.point import Point

from config import KEY_PATH
from core.utils import init_log
from avvyprotocol.l3_pb2 import AuthInfo, Signer, Signature

log = init_log()

current_curve = curve.secp256k1


def _make_keys():
    private_key = keys.gen_private_key(current_curve)
    public_key = keys.get_public_key(private_key, current_curve)
    return public_key, private_key


def sign_data(priv_key, data):
    r, s = ecdsa.sign(data, priv_key, hashfunc=sha256, curve=current_curve)
    return f'{r}:{s}'.encode()


def verify(auth_info: AuthInfo, txBody):
    verifies = []
    data = txBody.SerializeToString()

    for i, signer in enumerate(auth_info.signers):
        _signer = load_pub(signer.public_key)
        r, s = auth_info.signatures[i].data.split(b':')
        r, s = int(r), int(s)
        result = ecdsa.verify((r, s), data, _signer, hashfunc=sha256, curve=current_curve)
        verifies.append(result)

    return all(verifies)


def prepare_keys(key_path=KEY_PATH):
    """
    Загрузка ключей из фс. Либо их создание.
    """
    try:
        priv_key, pub_key = keys.import_key(key_path, curve=current_curve)
    except Exception as e:
        log.warning(e)
        log.warning(f'Make new key')
        priv_key = keys.gen_private_key(current_curve)
        pub_key = keys.get_public_key(priv_key, current_curve)
        keys.export_key(priv_key, filepath=key_path, curve=current_curve)

    return priv_key, pub_key


def load_pub(raw_data: bytes):
    x, y = tuple([int(el.decode()) for el in raw_data.split(b'-')])
    return Point(x, y, curve=current_curve)


def dump_pub(pub) -> bytes:
    return f'{pub.x}-{pub.y}'.encode()


class TypedSingleton(type):
    instances = {}
    key = lambda cls, args, kwargs: f'{cls}{args}{kwargs}'

    def __call__(cls, *args, **kwargs):
        key = TypedSingleton.key(cls, args, kwargs)
        if key not in TypedSingleton.instances:
            TypedSingleton.instances[key] = super(TypedSingleton, cls).__call__(*args, **kwargs)
        return TypedSingleton.instances[key]


class SigningService(metaclass=TypedSingleton):
    """
    Предоставляет подписание сериaлизуемых данных.
    # todo сделать нормальный сервис подписи на уровне с хранилищем ключей
    """

    def __init__(self, f_path=None):
        self.__priv, self.__pub = prepare_keys(key_path=f_path or KEY_PATH)

    @property
    def pub(self):
        return dump_pub(self.__pub)

    def get_auth_info(self, body):
        return AuthInfo(
            signers=[Signer(
                public_key=self.pub
            )],
            signatures=[
                Signature(
                    data=sign_data(self.__priv, body.SerializeToString())
                )
            ]
        )
