import json
import os

import requests
from aiohttp import ClientSession
from avvyprotocol.l2_pb2 import Rollup
from avvyprotocol.l2_pb2_grpc import L2ServiceStub
from grpc import insecure_channel

from config import L2_GRPC_HOST, L2_GRPC_PORT, IIDAP_ID
from core.utils import init_log
from core.verify import SigningService

log = init_log('l2')


class L2ClientInterface:
    signing_service = SigningService()

    def l3_register(self):
        pass

    async def send_rollup_async(self, rollup: Rollup):
        raise NotImplementedError

    def send_rollup(self, rollup: Rollup):
        log.info(f': {rollup.prev_hash} -> {rollup.current_hash} [{len(rollup.txns_list)}]')

    def attach_user(self, user_account, user_name):
        pass

    def detach_user(self, user_account):
        pass

    def get_last_accepted_rollup_digest(self):
        """
        Вернет либо последний утвержденный роллап, либо взорвется если ошибка,
        либо None в случае первого запуска иидапа
        """
        log.info(f'... get last rollup digest for iidap_id = {IIDAP_ID}')
        return None

    def set_rollup_digest_status(self, digest_id, rollup, status):
        """
        Установка статуса роллапа
        """
        log.info(f'... set status for rollup')
        return None


class GRPCL2Client(L2ClientInterface):
    def __init__(self):
        channel = insecure_channel(f'{L2_GRPC_HOST}:{L2_GRPC_PORT}')
        self.stub = L2ServiceStub(channel)

    def l3_register(self):
        pass

    def send_rollup(self, rollup: Rollup):
        pass


blockchain_base_url = os.environ.get('L2_API_URL', 'http://0.0.0.0:1317')
tendermint_node_url = os.environ.get('L2_TENDERMINT_URL', 'http://0.0.0.0:26657')


async def receive_events(message_action, operator='='):
    """
    Слушает вебсокет на наличие определенных событий в космосе. Оператор может быть CONTAINS либо =
    Экшены: (Create|Update|Delete)(Rollup|RollupDigest|...)
    """
    url = f'{tendermint_node_url}/websocket'
    try:
        async with ClientSession().ws_connect(url) as ws:
            await ws.send_json({
                'jsonrpc': '2.0',
                'method': 'subscribe',
                'id': 0,
                'params': {'query': f"tm.event = 'Tx' AND message.action {operator} '{message_action}'"}
            })
            async for msg in ws:
                log.info(f'msg: {msg}')
                data = json.loads(msg.data)
                result = data.get('result')
                if result:
                    yield result
    except Exception as e:
        log.exception(e)


def get_items(item_name, *args):
    """
    Возвращает список обьектов космоса по имени. В аргументах можно перечислить ключи маэппинга .
    Итемы: camelcase with capital. rollup, rollup_digest, attached_user, ...
    # todo сделать итератор с пагинацией
    """
    url = f"{blockchain_base_url}/avvy_l2/ltwonode/ltwonode/{item_name}"
    if args:
        url += '/' + '/'.join(args)
    response = requests.get(url)
    if response and response.status_code == 200:
        data = response.json()
        _item_name = ''.join(item_name.split('_'))
        for k, v in data.items():
            if k.lower() == _item_name.lower():
                return v
