import asyncio
import concurrent.futures
import hashlib
import os
import subprocess
import time

import yaml
from avvyprotocol.l2_pb2 import Rollup

from config import IIDAP_ID
from core.utils import init_log
from l2.client import L2ClientInterface, get_items
from l2.utils import parse_event_data

log = init_log('l2')
cli_exec = os.environ.get('L2_CLI_EXEC', '/home/alesha/go/bin/lTwoNoded tx ltwonode')


def hash(data):
    f = hashlib.md5()
    f.update(data.encode())
    r = int.from_bytes(f.digest(), 'big')
    return r >> r.bit_length() - 32


class SubProcL2Client(L2ClientInterface):
    def __init__(self):
        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=3)

    def l3_register(self):
        result = _process_command(_cmd_register(self.signing_service.pub))
        log.info(f'L3 registered: {result}')
        return result

    async def send_rollup_async(self, rollup):
        loop = asyncio.get_event_loop()
        tsk = loop.run_in_executor(self._executor, self.send_rollup, rollup)
        d, p = await asyncio.wait([tsk])
        if d:
            res = d.pop()
            if not res.exception():
                return int(res.result())
            else:
                raise res.exception()

    def send_rollup(self, rollup: Rollup):
        """
        Send rollup to cosmos. Make rollup digest, with ON_CHECK status
        """
        result = _process_command(_cmd_send_rollup(rollup))
        log.info(f'{rollup.current_hash} sent: {result}')
        digest_result = _process_command(_cmd_create_rollup_digest(IIDAP_ID, rollup))
        if not digest_result:
            log.warning(f'{rollup.current_hash} digest result is none :(')
            return

        log.info(f'Rollup digest result: {digest_result}')
        event = parse_event_data(digest_result['events'])
        return event['RD_ID']

    def get_last_accepted_rollup_digest(self):
        digests = get_items('rollup_digest')
        for d in reversed(digests):
            if d['status'] == 'ACCEPTED':
                return d

    def set_rollup_digest_status(self, digest_id, rollup, status):
        result = _process_command(_cmd_update_rollup_digest_status(digest_id, rollup, status, IIDAP_ID))
        event = parse_event_data(result['events'])
        return event['RD_ID']

    def attach_user(self, user_account, user_name):
        result = _process_command(_cmd_attach_user(IIDAP_ID, user_account, user_name))
        log.info(f'Attach user result: {result}')
        return result

    def detach_user(self, user_account):
        result = _process_command(_cmd_detach_user(IIDAP_ID, user_account))
        log.info(f'Detach user result: {result}')
        return result

    def check_user(self, user_account):
        return get_items('attachedUser', IIDAP_ID, user_account)


def _cmd_register(pub, name=None):
    return [f'{cli_exec} create-l-3-node "{pub.decode()}" "{name or "unnamed_l3_node"}" --from alice -y', ]


def _cmd_send_rollup(rollup):
    _txn_list = ' '.join([f'"{txn.SerializeToString()}"' for txn in rollup.txns_list])
    return [
        f'{cli_exec} create-rollup "{rollup.prev_hash.decode()}" "{rollup.current_hash.decode()}" {rollup.data_source_uri.decode()}'
        f' TXNS_LIST --from alice -y', ]


def _cmd_create_rollup_digest(iidap_id, rollup):
    return [
        f'{cli_exec} create-rollup-digest "{rollup.data_source_uri.decode()}" "{rollup.current_hash.decode()}" "ON_CHECK" "{iidap_id}" --from alice -y']  # todo make normal status model


def _cmd_update_rollup_digest_status(digest_id, rollup, status, iidap_id):
    return [
        f'{cli_exec} update-rollup-digest {digest_id} "{rollup.data_source_uri.decode()}" "{rollup.current_hash.decode()}" "{status}" "{iidap_id}" --from alice -y']


def _cmd_attach_user(iidap_id, user_account, user_name):
    return [f'{cli_exec} create-attached-user {iidap_id} {user_account} {user_name} --from alice -y']


def _cmd_detach_user(iidap_id, user_account):
    return [f'{cli_exec} delete-attached-user {iidap_id} {user_account} --from alice -y']


def _parse_result(result):
    result = yaml.load(result, Loader=yaml.FullLoader)
    result['code'] = int(result.get('code', -1))
    return result


def _run_cmd(cmd):
    log.info(f'run cmd: {cmd}')
    _start = time.time()
    with subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True) as p:
        p.wait(5)
        if p.returncode != 0:
            raise Exception(
                f'command: {cmd}\n return code: {p.returncode}\nproc: {p}'
            )

        result = p.stdout.read()
        log.info(f'cmd ok! result: {result} time: {time.time() - _start}')
        return _parse_result(result.decode())


def _process_command(cmd):
    try:
        result = _run_cmd(cmd)
        return result
    except Exception as e:
        log.exception(e)
        return None


