import asyncio
import random
import signal
import time
from asyncio import sleep
from multiprocessing import Process

import yaml
from aiohttp import ClientSession
from avvyprotocol.l3_pb2 import MsgResponse

from client.connect import connect_directly
from client.generators import np_matrix, make_add_object_msg, make_txn_request
from config import CONSENSUS_NODES_FILE
from core.utils import generate_id, init_log
from core.verify import SigningService
from load_test.msg_store import MakeTestStore
from load_test.utils import set_timestamp

log = init_log('client')


def child(p):
    return {'object_id': generate_id(),
            'object_class': 0,
            'transform': np_matrix(random.random()),
            'parent': p}


def msg_body():
    object_id = generate_id()
    return {
        'object_id': object_id,
        'transform': np_matrix(random.random()),
        'view_id': random.randint(-10, 10),
        'type': random.choice(['block', 'struct', 'tree', 'test']),
        'parent': random.choice([None, generate_id(), 100500]),
        'children': [
            child(object_id), child(object_id), child(object_id)
        ]
    }


txn_count_per_second = 10
msg_count_per_txn = 2


class L3Client:
    def __init__(self, number, test_marker):
        self.number = number
        self.test_marker = test_marker
        self.signing_service = SigningService(f_path=f'./load_client_{number}')
        with open(CONSENSUS_NODES_FILE) as f:
            self.nodes = yaml.load(f.read(), Loader=yaml.FullLoader)['nodes']
        self.loop = asyncio.get_event_loop()
        self.msg_store = MakeTestStore()

    async def ws_receive(self, host, port):
        url = f'http://{host}:{port}/ws/{self.signing_service.pub.decode()}'
        async with ClientSession().ws_connect(url) as ws:
            log.info(f'WS {url} connected!')
            msgs = 0
            await ws.send_json({})
            async for msg in ws:
                msgs += 1
                if msgs % 100 == 0:
                    log.info(f'from {url} received {msgs} subscribed msg')
                if msg.data:
                    try:
                        resp = MsgResponse()
                        resp.ParseFromString(msg.data)
                        self.msg_store.add_processed_msg(self.test_marker, resp.msg_id, resp.body)
                    except Exception as e:
                        log.exception(e)

    async def grpc_communicate(self, host, port):
        async with connect_directly(f'{host}:{port}') as stub:
            count = 0
            while 1:
                txn = make_txn_request(msgs=[make_add_object_msg(msg_body()), make_add_object_msg(msg_body())],
                                       signing_service=self.signing_service)
                set_timestamp(txn, '_tst_client_send')
                txn_response = await stub.ProcessTxn(txn)
                if txn_response.body.responses[0].status == MsgResponse.Status.REDIRECT:
                    return txn_response.body.responses[0].body['redirect_uri'].decode()

                await sleep(1. / txn_count_per_second)
                count += 1
                if count % 100 == 0:
                    log.info(f'GRPC l3 client communicate {count} txns')

    def _get_node(self, redirect_uri):
        host, port = redirect_uri.split(':')
        port = int(port)
        for node in self.nodes:
            if host == node['host'] and port == node['grpc_port']:
                return node

    def start(self):
        def g_shtdwn(loop):
            loop.stop()

        self.loop.add_signal_handler(signal.SIGINT, g_shtdwn)

        node = random.choice(self.nodes)
        while 1:
            ws_rcv_task = self.loop.create_task(self.ws_receive(node['host'], node['ws_port']))
            redirect_uri = self.loop.run_until_complete(self.grpc_communicate(node['host'], node['grpc_port']))
            node = self._get_node(redirect_uri)
            ws_rcv_task.cancel()


client_counts = [1, 3, 5, 10, 30, 40, 50, 70, 80, 90, 100]


def make_test():
    def proc_target():
        L3Client(i, cc).start()

    for cc in client_counts:
        log.info(f'Start test with {cc} clients')
        procs = []
        for i in range(cc):
            p = Process(target=proc_target)
            p.start()
            procs.append(p)

        time.sleep(120)
        for p in procs:
            p.terminate()
            p.join()


if __name__ == '__main__':
    make_test()
