import asyncio
import io
import time
from unittest import TestCase
from unittest.async_case import IsolatedAsyncioTestCase

from client.generators import TimedTxnManager
from core.consensus import Consensus, NodesConfig
from core.subscribe.event_listeners import EventListeners
from core.transactions_processing import TxnsProcessor
from core.utils import init_log
from core.verify import SigningService

log = init_log(only_stream=True)

nodes_conf = lambda x: [
    (f'localhost:{x}666', [f'localhost:{x}667', f'localhost:{x}668', ]),
    (f'localhost:{x}667', [f'localhost:{x}666', f'localhost:{x}668', ]),
    (f'localhost:{x}668', [f'localhost:{x}667', f'localhost:{x}666', ]),
]
nodes_conf_count = 1


def get_next_consensus_nodes_cfg():
    global nodes_conf_count
    nodes_conf_count += 1
    return nodes_conf(nodes_conf_count)


def wait_consensuses_leader(cons_dict):
    def get_cons_leader(cons_dict):
        leaders = set([cons._getLeader() for cons in cons_dict.values()])
        if len(leaders) == 1 and len(leaders.difference([None])) == 1:
            return leaders.pop()

    leader = get_cons_leader(cons_dict)
    while not leader:
        time.sleep(0.5)
        leader = get_cons_leader(cons_dict)

    return leader.address


class FakeTxnsProcessor(TxnsProcessor):
    def __init__(self, loop, event_listeners: EventListeners):
        super().__init__(loop, event_listeners)
        self.proc_txns = []

    def add_txn_to_processing(self, txn):
        self.proc_txns.append(txn)


class ConsensusConfigTestCase(TestCase):
    def test_nodes_cfg(self):
        fdata = """
        nodes:
          - { host: node1, raft_port: 7770, grpc_port: 5050 }
          - { host: node2, raft_port: 7771, grpc_port: 5051 }
          - { host: node3, raft_port: 7772, grpc_port: 5052 }
            """.strip()
        f = io.StringIO(fdata)
        nodes_cfg = NodesConfig(cfg_file=f, current_node='node1:7770:5050')
        c, a = nodes_cfg.get_raft_connection_info()
        assert c == 'node1:7770'
        assert len(a) == 2

        gu = nodes_cfg.get_grpc_uri('node3:7772')
        assert gu == 'node3:5052'


class ConsensusTestCase(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.e_listeners = EventListeners()
        self.nodes = get_next_consensus_nodes_cfg()
        self.txns_procs = {
            nc[0]: FakeTxnsProcessor(asyncio.get_event_loop(), event_listeners=self.e_listeners)
            for nc in self.nodes
        }
        self.consensuses = {
            nc[0]: Consensus(txns_proc=self.txns_procs[nc[0]], current_node=nc[0], other_nodes=nc[1])
            for nc in self.nodes
        }

    def tearDown(self) -> None:
        for _, c in self.consensuses.items():
            c.destroy()
        del self.consensuses

    async def test_smoke(self):
        leader = wait_consensuses_leader(self.consensuses)
        ttm = TimedTxnManager(SigningService('test_key'))

        for tx in ttm.generate_txs(1, 10):
            self.consensuses[leader].sync_transaction(tx)

        while not all([cns.sync_proc_debt == 0 for cns in self.consensuses.values()]):
            time.sleep(.5)

        assert all([len(proc.proc_txns) == 1 for proc in self.txns_procs.values()])

        for tx in ttm.generate_txs(1, 10):
            self.consensuses[leader].sync_transaction(tx)

        while not all([cns.sync_proc_debt == 0 for cns in self.consensuses.values()]):
            time.sleep(.5)

        assert all([len(proc.proc_txns) == 2 for proc in self.txns_procs.values()])

        assert self.consensuses[leader]._isLeader()