import unittest
from asyncio import sleep

from avvyprotocol.l2_pb2 import Rollup

from client.generators import TimedTxnManager
from config import ROLLUP_MAX_TXN_COUNT, IIDAP_ID, TICK_FRAME_RATE_S, SHUTDOWN_WAIT_TICK_COUNT
from core.consensus import Consensus
from core.rollup import RollupProcessor, RollupDigestProcessor
from core.startup import StartEngine
from core.store.prove import Storage
from core.subscribe.event_listeners import EventListeners
from core.utils import generate_id
from core.verify import SigningService
from l2.client import L2ClientInterface
from l2.subproc import SubProcL2Client


class FakeConsensus(Consensus):
    synced_txns = []
    leader = None

    def __init__(self):
        pass

    def _isLeader(self):
        return self.leader == None


class FakeL2Client(L2ClientInterface):
    sent_rollups = []

    def __init__(self):
        pass

    def send_rollup(self, rollup: Rollup):
        self.sent_rollups.append(rollup)

    async def send_rollup_async(self, rollup: Rollup):
        self.sent_rollups.append(rollup)


class FakeStartEngine(StartEngine):
    starts = 0
    shutdowns = 0

    def start_iidap(self):
        self.starts += 1

    def shutdown_iidap(self):
        self.shutdowns += 1

    def event_restart(self):
        self.starts += 1

    def event_shutdown(self):
        self.shutdowns += 1


class FakeEventListeners(EventListeners):
    msgs = []

    async def broadcast_msg(self, msg):
        self.msgs.append(msg)


class RollupTest(unittest.IsolatedAsyncioTestCase):
    async def test_rollup(self):
        storage = Storage(IIDAP_ID)
        event_listeners = FakeEventListeners()
        start_engine = FakeStartEngine()
        consensus = FakeConsensus()
        l2_client = FakeL2Client()
        rollup_proc = RollupProcessor(storage=storage,
                                      event_listeners=event_listeners,
                                      start_engine=start_engine,
                                      consensus=consensus,
                                      l2_client=l2_client)

        ttm = TimedTxnManager(SigningService('test_key'))
        for txn in ttm.generate_txs(ROLLUP_MAX_TXN_COUNT * 3, 1):
            await rollup_proc.append_txn(txn)

        assert len(l2_client.sent_rollups) == 3

        consensus.leader = 'some leader'
        for txn in ttm.generate_txs(ROLLUP_MAX_TXN_COUNT, 1):
            await rollup_proc.append_txn(txn)

        assert len(l2_client.sent_rollups) == 3
        await sleep(TICK_FRAME_RATE_S * SHUTDOWN_WAIT_TICK_COUNT + 1)
        assert rollup_proc.start_engine.shutdowns == 1

        rollup_proc.stop_task()

    async def test_rollup_digest(self):
        event_listeners = FakeEventListeners()
        start_engine = FakeStartEngine()
        # prepare
        rollup_digest_proc = RollupDigestProcessor(event_listeners, start_engine)

        # check not sent rollup digests
        digest_statuses = list(rollup_digest_proc._sent_rollups_digests.items())
        self.assertEqual(digest_statuses, [])

        l2_client = SubProcL2Client()
        # send rollup.
        rollup = Rollup(id=generate_id(), prev_hash=b'prev', current_hash=b'curr', txns_list=[], data_source_uri=b'source_uri')
        rollup_digit_id = await l2_client.send_rollup_async(rollup)
        self.assertIsNotNone(rollup_digit_id)
        await sleep(1)

        # after send rollup must have ON_CHECK rollup digest
        digest_statuses = list(rollup_digest_proc._sent_rollups_digests.items())
        self.assertEqual(digest_statuses, [(rollup_digit_id, 'ON_CHECK')])
        l2_client.set_rollup_digest_status(rollup_digit_id, rollup, 'DECLINED')
        await sleep(1)

        # after change must have DECLINED rollup digest and restarted
        digest_statuses = list(rollup_digest_proc._sent_rollups_digests.items())
        self.assertEqual(digest_statuses, [(rollup_digit_id, 'DECLINED')])
        self.assertEqual(rollup_digest_proc.start_engine.starts, 1)

        rollup_digest_proc.stop_task()
