import unittest
from unittest.async_case import IsolatedAsyncioTestCase

from client.generators import np_array, make_msg
from core.subscribe.event_listeners import EventListeners
from core.subscribe.filters import SubscribeFilter
from core.utils import generate_id, dumps
from avvyprotocol.l3_pb2 import Signer, Msg, Event, MsgResponse
from ws.connection import ClientConnection


class SubscribeTestCase(unittest.TestCase):
    def test_smoke(self):
        s = SubscribeFilter(
            make_msg({'subscribe_filters': [
                {'prop_name_1': [{'between': [2, 4]}]},
                {'prop_name_2': [{'>': 0}]}
            ]}, type='subscribe')
        )

        assert s.filter_data({
            'prop_name_1': 2,  # ok
            'prop_name_2': 1  # ok
        })

        assert s.filter_data({
            'prop_name_1': 1,
            'prop_name_2': 1  # ok
        })

        assert not s.filter_data({
            'prop_name_1': 1,
            'prop_name_2': 0,
        })

        assert not s.filter_data({
            'prop_name_1': 'test',
            'prop_name_2': 'foo',
        })

    def test_np_array(self):
        s = SubscribeFilter(
            make_msg({'subscribe_filters': [
                {'prop_name_1': [{'between': [np_array(0), np_array(2)]}]},
                {'prop_name_2': [{'>': np_array(1)}]}
            ]}, type='subscribe')
        )

        assert s.filter_data({
            'prop_name_1': np_array(1),  # ok
            'prop_name_2': np_array(2)  # ok
        })

        assert s.filter_data({
            'prop_name_1': np_array(-1),
            'prop_name_2': np_array(2)  # ok
        })

        assert not s.filter_data({
            'prop_name_1': np_array(3),
            'prop_name_2': np_array(0),
        })

        assert not s.filter_data({
            'prop_name_1': 'test',
            'prop_name_2': 'foo',
        })

        assert not s.filter_data({
            'prop_name_1': 1,
            'prop_name_2': 2,
        })


class FakeClientConnection(ClientConnection):
    def __init__(self):
        super().__init__(None)
        self.messages = []

    async def handle_connection(self):
        return await super().handle_connection()

    async def send_message(self, msg):
        self.messages.append(msg)

    async def close_connection(self, ws_close_code=None):
        return await super().close_connection(ws_close_code)


class TestEventListeners(IsolatedAsyncioTestCase):
    async def test_smoke(self):
        e_listeners = EventListeners()

        msgs = [Msg(id=generate_id(), body={'field': str(i).encode()}) for i in range(10)]
        msg_responses = [MsgResponse(msg_id=msg.id, status=MsgResponse.Status.ACCEPTED) for msg in msgs]

        conn1 = FakeClientConnection()
        conn2 = FakeClientConnection()
        client1 = Signer(public_key=b'test1')
        client2 = Signer(public_key=b'test2')

        client1_id = client1.public_key.decode()
        client2_id = client2.public_key.decode()
        e_listeners.set_client_connection(client1_id, conn1)
        e_listeners.set_client_connection(client2_id, conn2)

        await e_listeners.broadcast_subscribed_msg_events([Event(initiator=client1, msg=msgs[0])])
        assert len(conn1.messages) == 1
        assert conn1.messages[0].id == msgs[0].id
        assert conn2.messages == []

        await e_listeners.broadcast_subscribed_msg_events([Event(initiator=client2, msg=msgs[1])])
        assert len(conn1.messages) == 1
        assert conn1.messages[0].id == msgs[0].id
        assert len(conn2.messages) == 1
        assert conn2.messages[0].id == msgs[1].id

        await e_listeners.send_directly_responses(client1_id, msg_responses[:2])
        assert len(conn1.messages) == 3
        assert len(conn2.messages) == 1

        await e_listeners.send_directly_responses(client2_id, msg_responses[:2])
        assert len(conn1.messages) == 3
        assert len(conn2.messages) == 3

        e_listeners.set_client_subscribe_filter(
            client1_id,
            SubscribeFilter(
                make_msg({'subscribe_filters': [{'field': [{'=': 1}]}]}, type='subscribe'))
        )
        await e_listeners.broadcast_subscribed_msg_events([Event(initiator=client2, msg=msgs[1])])
        assert len(conn1.messages) == 4
        assert len(conn2.messages) == 4
        assert conn1.messages[-1].id == msgs[1].id
        assert conn2.messages[-1].id == msgs[1].id

        await e_listeners.broadcast_subscribed_msg_events([Event(initiator=client2, msg=msgs[2])])
        assert len(conn1.messages) == 4
        assert len(conn2.messages) == 5
        assert conn1.messages[-1].id == msgs[1].id
        assert conn2.messages[-1].id == msgs[2].id
