package compiler

import (
	"errors"

	"golang.org/x/exp/utf8string"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang"
	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"
)

const (
	GramBlockBegin     = '{'
	GramBlockEnd       = '}'
	GramListBegin      = '['
	GramListEnd        = ']'
	GramString         = '"'
	GramLibRefPrefix   = '$'
	GramInputRefPrefix = '@'
	GramRefAt          = '.'
	GramRefRoot        = '~'
	GramEndLine        = ';'
)

var (
	ErrUnkownSymbol   = errors.New("UNKOWN SYMBOL")
	ErrNotImplemented = errors.New("NOT IMPLEMENTED")
	ErrBlockNotFound  = errors.New("BLOCK NOT FOUND")
	ErrNameNotFound   = errors.New("NAME NOT FOUND")
	ErrValueNotFound  = errors.New("VALUE NOT FOUND")
	ErrStringNotFound = errors.New("STRING NOT FOUND")
	ErrListNotFound   = errors.New("LIST NOT FOUND")
)

func ParseBlock(s *utf8string.String) (props []*aproto.Prop, end int, err error) {
	props = make([]*aproto.Prop, 0)

	var r int
	var done bool

	var begined bool
	var ended bool

	for r < s.RuneCount() && !done && err == nil {
		cr := s.At(r)
		switch cr {
		case ' ', '\t', '\n', GramEndLine:
			r++
		case GramBlockBegin:
			begined = true
			r++
		case GramBlockEnd:
			end = r + 1
			ended = begined
			done = true
		default:
			if isAlpha(cr) {
				var (
					prop *aproto.Prop
				)
				prop, end, err = ParseProp(utf8string.NewString(s.Slice(r, s.RuneCount())))
				if err == nil {
					r += end
					end = r
					props = append(props, prop)
				}
			} else {
				err = ErrUnkownSymbol
			}
		}
	}

	if err == nil && !ended {
		err = ErrBlockNotFound
	}

	return
}

func ParseProp(s *utf8string.String) (prop *aproto.Prop, end int, err error) {
	prop = &aproto.Prop{}

	var nameEnd int
	prop.Name, nameEnd, err = ParseName(s)
	if err != nil {
		return
	}
	prop.Value, end, err = ParseValue(utf8string.NewString(s.Slice(nameEnd, s.RuneCount())))
	end += nameEnd

	return
}

func ParseName(s *utf8string.String) (name string, end int, err error) {
	var l, r int
	var done bool

	for r < s.RuneCount() && !done && err == nil {
		cl := s.At(l)
		cr := s.At(r)
		switch cr {
		case ' ', '\t', '\n':
			if done = !isSpace(cl); !done {
				r++
			}
		case ':', ';':
			done = true
		default:
			if isSpace(cl) {
				l = r
			}
			r++
		}
	}

	if l == r {
		err = ErrNameNotFound
	} else {
		name = s.Slice(l, r)
		end = r
	}

	return
}

func ParseValue(s *utf8string.String) (value *aproto.Payload, end int, err error) {

	var r int
	var done bool

	var ref *aproto.Reference
	var block []*aproto.Prop

	for r < s.RuneCount() && !done && err == nil {
		cr := s.At(r)
		switch cr {
		case ' ', '\t', '\n', ':':
			r++
		case GramEndLine:
			end = r + 1
			done = true
		case GramBlockBegin:
			block, end, err = ParseBlock(utf8string.NewString(s.Slice(r, s.RuneCount())))
			if err == nil {
				end += r
			}
			done = true
		case GramListBegin:
			var list []*aproto.Payload
			list, end, err = ParseList(utf8string.NewString(s.Slice(r, s.RuneCount())))
			if err == nil {
				end += r
				value = lang.ListPayloadPayload(list, nil)
			}
			done = true
		case GramString:
			var str string
			str, end, err = ParseString(utf8string.NewString(s.Slice(r, s.RuneCount())))
			if err == nil {
				end += r
				value = lang.StringPayload(str)
			}
			done = true
		default:
			if isAlpha(cr) || cr == GramRefAt || cr == GramRefRoot || cr == GramLibRefPrefix || cr == GramInputRefPrefix {
				ref, end, err = ParseRef(utf8string.NewString(s.Slice(r, s.RuneCount())))
				if err == nil {
					r += end
					end = r
				}
			} else if isDigit(cr) {
				err = ErrNotImplemented
			} else if ref != nil {
				done = true
			} else {
				err = ErrUnkownSymbol
			}
		}
	}

	if err != nil {
		return
	}

	if value == nil {
		if ref != nil {
			if block != nil {
				value = lang.ClosurePayload(&aproto.Closure{
					Ref:     ref,
					Context: block,
				})
			} else {
				value = lang.RefPayload(ref)
			}
		} else if block != nil {
			value = lang.PropsPayload(block)
		}
	}

	if value == nil {
		err = ErrValueNotFound
	}

	return
}

func ParseRef(s *utf8string.String) (ref *aproto.Reference, end int, err error) {
	name, end, err := ParseName(s)
	if err != nil {
		return
	}

	ref = &aproto.Reference{}
	switch rune(name[0]) {
	case GramLibRefPrefix:
		ref.Context = aproto.Reference_BUILT_IN
		ref.Name = name[1:]
	case GramInputRefPrefix:
		ref.Context = aproto.Reference_INPUT
		ref.Name = name[1:]
	default:
		ref.Context = aproto.Reference_LOCAL
		ref.Name = name
	}

	return
}

func ParseString(s *utf8string.String) (str string, end int, err error) {

	var l, r int
	var done bool

	for r < s.RuneCount() && !done && err == nil {
		cr := s.At(r)
		switch cr {
		case GramString:
			if l != r {
				if s.At(l) != GramString {
					l = r
				} else {
					done = true
				}
			} else {
				r++
			}
		default:
			r++
		}
	}

	cl := s.At(l)
	cr := s.At(r)
	if l == r || cl != GramString || cr != GramString {
		err = ErrStringNotFound
	} else {
		str = s.Slice(l+1, r)
		end = r + 1
	}

	return
}

func ParseList(s *utf8string.String) (list []*aproto.Payload, end int, err error) {
	var r int
	var done bool

	var payload *aproto.Payload

	var begined bool
	var ended bool

	for r < s.RuneCount() && !done && err == nil {
		cr := s.At(r)
		switch cr {
		case ' ', '\t', '\n', ',':
			r += 1
		case GramListEnd:
			end = r + 1
			if begined {
				ended = true
			}
			done = true
		default:
			if !begined {
				if cr == GramListBegin {
					begined = true
					r += 1
				} else {
					err = ErrUnkownSymbol
				}
			} else {
				payload, end, err = ParseValue(utf8string.NewString(s.Slice(r, s.RuneCount())))
				if err == nil {
					r += end
					end = r

					list = append(list, payload)
				}
			}
		}
	}

	if err == nil && !ended {
		err = ErrListNotFound
	}

	return
}

func isAlpha(c rune) bool {
	return ('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z')
}

func isDigit(c rune) bool {
	return '0' <= c && c <= '9'
}

func isSpace(c rune) bool {
	switch c {
	case ' ', '\t', '\n':
		return true
	default:
		return false
	}
}
