package grid

import (
	"errors"
	"math"
	// "gonum.org/v1/gonum/mat"
)

type Point3D [3]int64
type BBox struct {
	MinPoint Point3D
	MaxPoint Point3D
}

func (bboxA *BBox) Join(bboxes ...*BBox) (newBBox *BBox) {
	newBBox = new(BBox)
	*newBBox = *bboxA
	for i := 0; i < 3; i++ {
		for _, bboxB := range bboxes {
			if bboxB.MinPoint[i] < newBBox.MinPoint[i] {
				newBBox.MinPoint[i] = bboxB.MinPoint[i]
			}
			if newBBox.MaxPoint[i] < bboxB.MaxPoint[i] {
				newBBox.MaxPoint[i] = bboxB.MaxPoint[i]
			}
		}
	}
	return
}

func (bboxA *BBox) Intersect(bboxB *BBox) (ok bool) {
	for i := 0; i < 3; i++ {
		ok = bboxA.MinPoint[i] <= bboxB.MaxPoint[i] && bboxB.MaxPoint[i] <= bboxA.MaxPoint[i] ||
			(bboxB.MinPoint[i] <= bboxA.MaxPoint[i] && bboxA.MaxPoint[i] <= bboxB.MaxPoint[i])
		if !ok {
			break
		}
	}
	return
}

const trMatrixSize = 4

// TransformMatrix
type TrMatrix [trMatrixSize][trMatrixSize]float64

// CONSTRUCTORS

func NewTrMatrix() *TrMatrix {
	return new(TrMatrix)
}

func NewEyeTrMatrix() (m *TrMatrix) {
	m = new(TrMatrix)
	for i := 0; i < 4; i++ {
		m[i][i] = 1
	}
	return
}

func NewEyeTrMatrixWithCoords(x, y, z float64) (m *TrMatrix) {
	m = NewEyeTrMatrix()
	m[3][0] = x
	m[3][0] = y
	m[3][0] = z
	return
}

func MakeTrMatrixFromFloats(nums []float64) (m *TrMatrix, err error) {
	if len(nums) != 16 {
		err = errors.New("INVALID SIZE OF FLOAT VECTOR")
		return
	}

	m = new(TrMatrix)

	numInd := 0
	for i := 0; i < 4; i++ {
		for j := 0; j < 4; j++ {
			m[i][j] = nums[numInd]
			numInd++
		}
	}

	return
}

// METHODS

func (m *TrMatrix) GetCoords() (point Point3D) {
	point[0] = int64(math.Round(m[3][0]))
	point[1] = int64(math.Round(m[3][1]))
	point[2] = int64(math.Round(m[3][2]))
	return
}

func (m *TrMatrix) GetEulerCoords(degrees bool) (point Point3D) {
	sY := math.Sqrt(m[0][0]*m[0][0] + m[1][0]*m[1][0])
	singular := sY < 1e-6

	var x, y, z float64

	if !singular {
		x = math.Atan2(m[2][1], m[2][2])
		y = math.Atan2(-m[2][0], sY)
		z = math.Atan2(m[1][0], m[0][0])
	} else {
		x = math.Atan2(-m[1][2], m[1][1])
		y = math.Atan2(-m[2][0], sY)
		z = 0
	}

	if degrees {
		point[0] = int64(math.Round(x*(180/math.Pi))) % 360
		point[1] = int64(math.Round(y*(180/math.Pi))) % 360
		point[2] = int64(math.Round(z*(180/math.Pi))) % 360
	} else {
		point[0] = int64(x)
		point[1] = int64(y)
		point[2] = int64(z)
	}

	return
}

func (m *TrMatrix) ToFloats() (nums []float64) {
	for i := 0; i < trMatrixSize; i++ {
		for j := 0; j < trMatrixSize; j++ {
			nums = append(nums, m[i][j])
		}
	}
	return
}

func (m *TrMatrix) Dot(op *TrMatrix) *TrMatrix {
	newM := new(TrMatrix)
	for i := 0; i < trMatrixSize; i++ {
		for j := 0; j < trMatrixSize; j++ {
			for k := 0; k < trMatrixSize; k++ {
				newM[i][j] += m[i][k] * op[k][j]
			}
		}
	}
	return newM
}

func (m *TrMatrix) ApplyRotationMatrix(rotM *TrMatrix) *TrMatrix {
	copy := *rotM
	for i := 0; i < 3; i++ {
		copy[3][i] = 0
	}

	return m.Dot(&copy)
}

func (m *TrMatrix) ApplyTransform(rM *TrMatrix) *TrMatrix {
	return m.Dot(rM)
}
