package grid

import (
	"math"
	"testing"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/utils"
)

func TestGrid(t *testing.T) {

	t.Run("TransformationMatrix", func(t *testing.T) {
		a := utils.TestAsserts(t)

		{
			m := NewEyeTrMatrix()
			a.NotNil(m)
			a.Equal(float64(0), m[0][1])
			a.Equal(float64(0), m[3][2])
		}

		{
			m, err := MakeTrMatrixFromFloats([]float64{
				1, 2, 3, 4,
				1, 2, 3, 4,
				1, 2, 3, 4,
				1, 2, 3, 4,
			})

			a.NoError(err)
			a.NotNil(m)
			for i := 0; i < 4; i++ {
				a.Equal(float64(i+1), m[i][i])
			}

			coords := m.GetCoords()
			a.Equal(int64(1), coords[0])
			a.Equal(int64(2), coords[1])
			a.Equal(int64(3), coords[2])
		}

		{
			m, err := MakeTrMatrixFromFloats([]float64{
				1, 2, 3, 4,
				1, 2, 3, 4,
				1, 2, 3, 4,
				1, 2, 3, 4,
			})

			a.NoError(err)
			a.NotNil(m)
			for i := 0; i < 4; i++ {
				a.Equal(float64(i+1), m[i][i])
			}
		}

		{
			floats := NewTrMatrix().ToFloats()
			a.Req.Len(floats, 16)

			for i := 0; i < 16; i++ {
				a.Equal(float64(0), floats[i])
			}
		}

	})

	t.Run("Object", func(t *testing.T) {
		a := utils.TestAsserts(t)

		object, err := MakeObjectFromProtoMsgBody(types.MsgBody{
			"transformation": utils.ListFloatPayload([]float32{
				1, 0, 0, 0,
				0, 1, 0, 0,
				0, 0, 1, 0,
				1, 2, 3, 1,
			}),
		})
		a.Req.NoError(err)
		a.Req.NotNil(object)

		bbox := object.GetBBox()
		a.Req.Equal(
			BBox{
				MinPoint: Point3D{1, 2, 3},
				MaxPoint: Point3D{1, 2, 3},
			},
			*bbox,
		)

		childs := utils.MapPayload(types.MsgBody{
			"a": utils.MapPayload(types.MsgBody{
				"transformation": utils.ListFloatPayload([]float32{
					1, 0, 0, 0,
					0, 1, 0, 0,
					0, 0, 1, 0,
					1, 2, 7, 1,
				}),
			}),
			"b": utils.MapPayload(types.MsgBody{
				"transformation": utils.ListFloatPayload([]float32{
					1, 0, 0, 0,
					0, 1, 0, 0,
					0, 0, 1, 0,
					5, 6, 3, 1,
				}),
			}),
		})
		err = object.ExtractChildsFromProtoPayload(childs)
		a.Req.NoError(err)
		a.Req.Len(object.childs, 2)

		bbox = object.GetBBox()
		a.Req.Equal(
			BBox{
				MinPoint: Point3D{2, 4, 6},
				MaxPoint: Point3D{6, 8, 10},
			},
			*bbox,
		)

		cells := object.GetCells()
		a.Req.Len(cells, 2)
		// a.Req.Equal(Point3D{2, 4, 10}, cells[0])
		// a.Req.Equal(Point3D{6, 8, 6}, cells[1])
	})

	t.Run("Object rotate", func(t *testing.T) {
		a := utils.TestAsserts(t)

		// 45 degree rotate around x-axis
		θ := math.Pi / 4
		a.Equal(θ, θ)
		object, err := MakeObjectFromProtoMsgBody(types.MsgBody{
			"transformation": utils.ListFloatPayload([]float32{
				1, 0, 0, 0,
				0, float32(math.Cos(θ)), float32(-math.Sin(θ)), 0,
				0, float32(math.Sin(θ)), float32(math.Cos(θ)), 0,
				0, 0, 10, 1,
			}),
			"children": utils.MapPayload(types.MsgBody{
				"a": utils.MapPayload(types.MsgBody{
					"transformation": utils.ListFloatPayload([]float32{
						1, 0, 0, 0,
						0, 1, 0, 0,
						0, 0, 1, 0,
						0, 0, 0, 1,
					}),
				}),
				"b": utils.MapPayload(types.MsgBody{
					"transformation": utils.ListFloatPayload([]float32{
						1, 0, 0, 0,
						0, 1, 0, 0,
						0, 0, 1, 0,
						0, 0, 1, 1,
					}),
				}),
			}),
		})
		a.Req.NoError(err)
		a.Req.NotNil(object)
		a.Req.Len(object.childs, 2)

		bbox := object.GetBBox()
		a.Req.Equal(
			BBox{
				MinPoint: Point3D{0, 0, 10},
				MaxPoint: Point3D{0, 1, 11},
			},
			*bbox,
		)

		cells := object.GetCells()
		a.Req.Len(cells, 2)

		// a.Req.Equal(Point3D{0, 0, 10}, cells[0])
		// a.Req.Equal(Point3D{0, 1, 11}, cells[1])
	})
}
