package grid

import (
	"errors"

	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"
)

type Object struct {
	trMatrix        *TrMatrix
	updatedTrMatrix *TrMatrix
	childs          map[string]*Object
}

func MakeObjectFromProtoMsgBody(msgBody types.MsgBody) (obj *Object, err error) {
	if msgBody == nil {
		err = errors.New("MSG BODY NOT PROVIDED")
		return
	}

	trPayload, ok := msgBody[spec.Transformation.String()]
	if !ok {
		err = errors.New("TRANSFORMATION NOT PROVIDED")
		return
	}

	var trMatrix *TrMatrix

	switch val := trPayload.GetValue().(type) {
	case *aproto.Payload_ListFloat:
		trMatrix, err = MakeTrMatrixFromFloats(val.ListFloat.Value)
	default:
		err = errors.New("INVALID TYPE OF TRANSORMATION PROP")
	}

	if err != nil {
		return
	}

	obj = &Object{
		trMatrix: trMatrix,
		childs:   make(map[string]*Object),
	}

	if childsPayload, ok := msgBody[spec.Children.String()]; ok {
		err = obj.ExtractChildsFromProtoPayload(childsPayload)
	}

	return
}

// Methods

func (obj *Object) ExtractChildsFromProtoPayload(childsPayload *aproto.Payload) (err error) {
	childs := make(map[string]*Object)

	if childsMap, ok := childsPayload.GetValue().(*aproto.Payload_Map); ok {
		for childID, childProps := range childsMap.Map.GetValue() {
			if childMap, ok := childProps.GetValue().(*aproto.Payload_Map); ok {
				var childObj *Object
				childObj, err = MakeObjectFromProtoMsgBody(childMap.Map.GetValue())
				if err != nil {
					break
				}
				childs[childID] = childObj
			} else {
				err = errors.New("INVALID TYPE OF CHILD PROP")
				break
			}
		}
	} else {
		err = errors.New("INVALID TYPE OF CHILDS PROP")
	}

	if err == nil {
		obj.childs = childs
	}

	return
}

func (obj *Object) GetTrMatrix() *TrMatrix {
	if obj.updatedTrMatrix != nil {
		return obj.updatedTrMatrix
	}
	return obj.trMatrix
}

func (obj *Object) GetBBox() (bbox *BBox) {
	if len(obj.childs) == 0 {
		bbox = new(BBox)
		bbox.MinPoint = obj.GetTrMatrix().GetCoords()
		bbox.MaxPoint = bbox.MinPoint
	} else {
		for _, child := range obj.childs {
			if child.updatedTrMatrix == nil {
				child.updatedTrMatrix = child.trMatrix.ApplyTransform(obj.GetTrMatrix())
			}
			childBBox := child.GetBBox()
			if bbox == nil {
				bbox = childBBox
			} else {
				bbox = bbox.Join(childBBox)
			}
		}
	}
	return
}

// NOTE: order of points is indeterminate
func (obj *Object) GetCells() (cells []Point3D) {
	obj.getCells(&cells)
	return
}

func (obj *Object) getCells(cells *[]Point3D) {
	if len(obj.childs) == 0 {
		*cells = append(*cells, obj.GetTrMatrix().GetCoords())
	} else {
		for _, child := range obj.childs {
			if child.updatedTrMatrix == nil {
				child.updatedTrMatrix = child.trMatrix.ApplyTransform(obj.GetTrMatrix())
			}
			child.getCells(cells)
		}
	}
}
