# DSL and concepts for user defined handlers

## Entities

### 1. DIRECTIVE

Defines sections with different behavior and parameters at the root of the handler.  
Directives are built in and specified for each processing scenario (custom handler, subscribe filter).  

Common:
- `#def` - section for definition custom functions of execution runtime
- `#filter` - define condition for message data filter

Handler:
- `#handler_type` - define type of transactions to be processed by the handler
- `#required_props` - list of required message props in transaction  
- `#txns` - define return transactions to be added to TxnProcessor
- `#events` - define return events to be added to EventBus

### 2. FUNCTION

Functions are pure.   
Can be built in and user defined.  
There are built in high order functions.  

(?) Function has named parameters. Calling a function requires passing parameters by name:  
```
{"$some_func": {"param1": val1, "param2": val2}}
```

**Main categories of built in functions:**

| Category               | Functions                                           |
| ---------------------- | --------------------------------------------------- |
| Logic                  | `$eq, $lt, $gt, $not, $and, $or` and etc            |
| Arithmetic             | `$plus, $minus, $multi, $dev, $mod, $round` and etc |
| Sequence               | `$concat, $union, $inter, $diff`                    |
| Linear                 | `$dot, $diag, $det` and etc                         |
| High order/Aggregation | `$map, $accumulate`                                 |
| IO                     | `$get_object(id, prop_name)`                        |


## Notation

| Pattern   | Means                                             | Example                                                     |
| --------- | ------------------------------------------------- | ----------------------------------------------------------- |
| `#%NAME%` | Directive                                         | `{"#def": {}}`, `{"#handler_type": "abc"}`                  |
| `$%NAME%` | Function (on calling)                             | `{"$plus": {"a":2, "b":2}}`, `{"$sqrt": 25}`, `"$gen_uuid"` |
| `&%NAME%` | Function reference (on passing)                   | `{"$map": {"func": "&sqrt", "list": [16, 25, 36]}}`         |
| `@%NAME%` | Function arg (on access inside function body)     |                                                             |
| `.%NAME%` | Property of input transaction message (on access) |                                                             |

## Handler definition example
```json
{
    // Define custom functions and constants (function that returns constant)
    "#def": {
        "make_hello": {
            "args": {"name": "string"},
            "body": {"$concat": {"a": "Hello,", "b": "@name"}},
        },
        "get_transformation": {
            "args": {"object_id": "string"},
            "body": {"$get_object_prop": {"object_id": "@object_id", "prop": "transformation"}}
        },
        "current_transformation": {
            "body": {"$get_transformation": ".object_id"}
        }
    },
    // Handle transaction with type "add_object"
    "#handler_type": "offset_object",
    // Process only msgs in the specified area
    "#filter": {
        "$lt": {"prop": "offset", "val": [10, 7, 4]}
    },
    // Handler will work with "transformation" prop and should to declare this
    "#required_props": {
        "offset": "[3]float"
    },
    // RETURN
    // ========================
    "#output": { 
      "state": {
          "transformation": {"$plus": {
              "a": "$current_transformation",
              "b": {
                  "$accumulate": {
                      "func": "&concat",
                      "list": [
                        [
                            0, 0, 0, 0,
                            0, 0, 0, 0,
                            0, 0, 0, 0
                        ],
                        ".offset", 
                        [0]
                      ],
                  }
              }
          }
      },
      "msgs": [
          // Transaction to create a new object with an offset of 5 in x and -3 in y relative to the current object
          {
              "type": "add_object",
              "object_id": {"$gen_object_id": {}},
              "transformation": {"$plus": {
                  "a": "$current_transformation",
                  "b": [
                    0,  0, 0, 0,
                    0,  0, 0, 0,
                    0,  0, 0, 0,
                    5, -3, 0, 0
                  ]
              }}
          }
      ],
      "events": [
          // Event for the client
          {
              "home": "$get_client_key",
              "message": {"$make_hello": {"name": "$get_client_key"}} // "Hello, 181210f8f9c779c2" 
          }
      ]
    }
    // ========================
}
```