# Handler semantics
## Terms
* Object - common entity in the Avvyland. Object has an account, properties,  
* Role - an id that means group of actors 
* Anyone - an id (or account address) that means anyone id or account (constant max(uint256))
* NoOne - an id or account address that means no one id or account (constant uint256(0))
* Transaction - a combination of the messages (transaction body) and body signature 
* Message - a combination of handler name and set of properties
* Handler - a method for processing some type of message
* Property - a key-value pair

## Handlers
Any (parent) object may have the next types of handlers:
1. **"createObject"** - creates and add the child object into the parent object
2. **"transferObject"** - transfer the object from one parent object (usually user object) to the other parent object (usually brane) or vice versa  
3. **"modifyObject"** - call some object's method that usually modifies object's state
4. **"groupObjects"** - compose the objects into the new parent-children structure and add the structure into the parent object
5. **"explodeObject"** - explode the composed object (structure) to the separate objects, destroy the empty structure and add the objects into the parent object
6. **"cloneObject"** - creates clone of the object with some modified props and add it into the parent object
7. **"destroyObject"** - remove the object from the parent object and destroy it 
8. **"transferTokens"** - transfer tokens from the object's account to the other's account 


### Value types
* basic types:
    * byte
    * int64
    * uint64
    * int256
    * uint256
    * float64
    * ivector3 - int32\[3\]
    * ivector4 - int32\[4\]
    * vector3 - float32\[3\]
    * vector4 - float32\[4\]
    * matrix4 - float32\[4\]\[4\]
    * rgba(n1, n2) - float32\[n1]\[n2]\[4]
    * tensor3(n1, n2, n3) - float32\[n1\]\[n2\]\[n3\]
    * bytes - byte\[n\]
    * mesh - byte\[n]
    * string - char\[n\]
    * prop - (key: string, value:any)
    * props - {prop[]...}
* integer - one of int64, uint64, int256, uint256
* numerical - one of int64, int256, float64, vector3, matrix4, tensor3
* scalar - one of int64, int256, float64
* matrix - one of ivector3, ivector4, vector3, vector4, matrix4
* tensor - one of vector3, matrix4, rgba, tensor3
* binary - one of string, bytes
* any - one of basic types
* array - 1d-array of any basic type, e.g. int64[]
* logic - serialized logic object


### Common object props
* messageValidation: {logic: {message: msgProps, object: objProps}}
* createObjectHandler: {logic: {message: msgProps, object: objProps}}
* transferObjectHandler: {logic: {message: msgProps, object: objProps}}
* modifyObjectHandler: {logic: {message: msgProps, object: objProps}}
* groupObjectsHandler: {logic: {message: msgProps, object: objProps}}
* explodeObjectHandler: {logic: {message: msgProps, object: objProps}}
* cloneObjectHandler: {logic: {message: msgProps, object: objProps}}
* destroyObjectHandler: {logic: {message: msgProps, object: objProps}}
* transferTokensHandler: {logic: {message: msgProps, object: objProps}}
* roles: mapping(senderId => string[])
* balance: uint256  // *protected prop*
* blockcells: ivector3[] // *protected prop*
  * bcOrigin: ivector3 // *protected prop*


### Common compulsory message props
* sender: uint256 - transaction sender ID
* txnType: string - type of the transaction



### Logic
**Logic** is an abstract structure which defines a transition function and has a inputs, set children transition functions and outputs.
```json
{
  "name": "%LOGIC_NAME",  // for no-coding UI and self-documenting code
  "inputs": {
    "%SOURCE_ID": {
      "name": "%SOURCE_NAME", // for object's configuration or no-coding UI and for self-documenting code
      "valueType": "%VALUE_TYPE",
      "defaultValue": "%VALUE",
      "range": {"from": "%VALUE", "to": "%VALUE"}, // optional, valid only for scalar value types
      "fromList": ["%VALUE"] // optional, may be used for limiting valid values to the only ones contained in the list  
      "expose": "%VALUE" // optional bool, set to true if you want to configure it as the object's propertiy, default false 
    }
  },
  "children": [
  ],
  "outputs": {
    "%SINK_ID": {
      "name": "%SINK_NAME",
      "valueType": "%VALUE_TYPE",
      "value": "%VALUE"
    }
  }
}
```

In example, magValidationHandler can be set up as follows:
```json
{
  "msgValidationLogic": {
    "name": "Default message validation logic",
    "inputs": {
      "message": {
        "name": "Message props",
        "valueType": "props",
        "defaultValue": null
      },
      "object": {
        "name": "Object props",
        "valueType": "props",
        "defaultValue": null
      }
    },
    "children": [
    ],
    "outputs": {
        "name": "is incoming message valid",
        "valueType": "props",
        "value": {
          "isValid": {
            "name": "is incoming message valid",
            "valueType": "bool",
            "value": {"$eq":  ["$message.sender", "$object.owner"]}
          }
        }
    },
    "outputs": {
      "isValid": {
        "name": "is incoming message valid",
        "valueType": "bool",
        "value": {"$eq":  ["$message.sender", "$object.owner"]}
      }
    }
  }
}
```

The logic object can be instantiated as follows:
* named:
```json
{
  "instance1": {"$msgValidationLogic": "$senderId"} 
}
```
* unnamed:
```json
{"$msgValidationLogic": "$senderId"} 
```


###The embedded logic objects:

| logicId   | inputs                                                                                                                          | outputs                                  | definition                                                                          | notes                         |
|-----------|---------------------------------------------------------------------------------------------------------------------------------|------------------------------------------|-------------------------------------------------------------------------------------|-------------------------------|
| negate    | x: numerical                                                                                                                    | y: numerical                             | {"$negate": x}                                                                      | element-wise for tensor types |
| add       | x: numerical[]                                                                                                                  | y: numerical                             | {"$add": x}                                                                         | element-wise for tensor types |
| abs       | x: numerical                                                                                                                    | y: numerical                             | {"$abs": x}                                                                         | element-wise for tensor types |
| acos      | x: numerical                                                                                                                    | y: numerical                             | {"$acos": x}                                                                        | element-wise for tensor types |
| acosh     | x: numerical                                                                                                                    | y: numerical                             | {"$acosh": x}                                                                       | element-wise for tensor types |
| all       | x: bool[]                                                                                                                       | y: bool                                  | {"$all": x}                                                                         |                               |
| any       | x: bool[]                                                                                                                       | y: bool                                  | {"$any": x}                                                                         |                               |
| and       | x: bool[]                                                                                                                       | y: bool                                  | {"$and": x}                                                                         |                               |
| or        | x: bool[]                                                                                                                       | y: bool                                  | {"$or": x}                                                                          |                               |
| at        | x: any[], ind: int64<br/> x: tensor3, ind: ivector3                                                                             | y: any                                   | {"$at": \[x, index\]}                                                               |                               |
| set       | x: any[], ind: int64, value: any<br/> x: tensor3, ind: ivector3, value: float64                                                 | y: any                                   | {"$at": \[x, index\]}                                                               |                               |
| asin      | x: numerical                                                                                                                    | y: numerical                             | {"$asin": x}                                                                        | element-wise for tensor types |
| asinh     | x: numerical                                                                                                                    | y: numerical                             | {"$asinh": x}                                                                       | element-wise for tensor types |
| atan      | x: numerical                                                                                                                    | y: numerical                             | {"$atan": x}                                                                        | element-wise for tensor types |
| atan2     | x1: numerical, x2: numerical                                                                                                    | y: numerical                             | {"$atan2": \[x1, x2\]}                                                              | element-wise for tensor types |
| atanh     | x: numerical                                                                                                                    | y: numerical                             | {"$atanh": x}                                                                       | element-wise for tensor types |
| len       | x: binary                                                                                                                       | y: numerical                             | {"$len": x}                                                                         | element-wise for tensor types |
| gasItems  | x: logic                                                                                                                        | y: uint256                               | {"$gasItems": x}                                                                    | element-wise for tensor types |
| ceil      | x: numerical                                                                                                                    | y: numerical                             | {"$ceil": x}                                                                        | element-wise for tensor types |
| concat    | x: binary[]<br/>x: array[]                                                                                                      | y: binary<br/>y: array                   | {"$concat": x}                                                                      |                               |
| merge     | x: tensor[], axis: uint64                                                                                                       | y: numerical                             | {"$merge": \[x, axis\]}<br/>{"$merge": {"x": x, "axis": axis}}                      |                               |
| cond      | if: bool, then: any, else: any                                                                                                  | y: numerical                             | {"$cond": \[if, then, else\]}<br/>{"$cond": {"if": if, "then": then, "else": else}} |                               |
| cast      | x: any, type: string                                                                                                            | y: any                                   | {"$cast": \[x, type\]}<br/>{"$cast": {"x": x, "type": type}}                        | element-wise for tensor types |
| ivector3  | x1: int64, x2: int64, x3: int64                                                                                                 | y: ivector3                              | {"ivector3": \[x1, x2, x3\]}                                                        | element-wise for tensor types |
| vector3   | x1: float64, x2: float64, x3: float64                                                                                           | y: vector3                               | {"ivector3": \[x1, x2, x3\]}                                                        | element-wise for tensor types |
| ivector4  | x1: int64, x2: int64, x3: int64, x4: int64                                                                                      | y: ivector4                              | {"ivector3": \[x1, x2, x3\]}                                                        | element-wise for tensor types |
| vector4   | x1: float64, x2: float64, x3: float64, x4: float64                                                                              | y: vector4                               | {"ivector3": \[x1, x2, x3\]}                                                        | element-wise for tensor types |
| cos       | x: numerical                                                                                                                    | y: numerical                             | {"$cos": x}                                                                         | element-wise for tensor types |
| sin       | x: numerical                                                                                                                    | y: numerical                             | {"$sin": x}                                                                         | element-wise for tensor types |
| cosh      | x: numerical                                                                                                                    | y: numerical                             | {"$cosh": x}                                                                        | element-wise for tensor types |
| der2rad   | x: numerical                                                                                                                    | y: numerical                             | {"$deg2rad": x}                                                                     | element-wise for tensor types |
| divide    | x1: numerical, x2: numerical                                                                                                    | y: numerical                             | {"$divide": \[x1, x2\]}                                                             | element-wise for tensor types |
| eq        | x1: any, x2: any                                                                                                                | y: bool                                  | {"$eq": \[x1, x2\]}                                                                 |                               |
| gt        | x1: scalar, x2: scalar                                                                                                          | y: bool                                  | {"$gt": \[x1, x2\]}                                                                 |                               |
| gte       | x1: scalar, x2: scalar                                                                                                          | y: bool                                  | {"$gte": \[x1, x2\]}                                                                |                               |
| lt        | x1: scalar, x2: scalar                                                                                                          | y: bool                                  | {"$lt": \[x1, x2\]}                                                                 |                               |
| lte       | x1: scalar, x2: scalar                                                                                                          | y: bool                                  | {"$lte": \[x1, x2\]}                                                                |                               |
| exp       | x: numerical                                                                                                                    | y: numerical                             | {"$exp": x}                                                                         | element-wise for tensor types |
| cosh      | x: numerical                                                                                                                    | y: numerical                             | {"$cosh": x}                                                                        | element-wise for tensor types |
| ln        | x: numerical                                                                                                                    | y: numerical                             | {"$ln": x}                                                                          | element-wise for tensor types |
| log10     | x: numerical                                                                                                                    | y: numerical                             | {"$log10": x}                                                                       | element-wise for tensor types |
| log       | x: numerical, base: scalar                                                                                                      | y: numerical                             | {"$log": \[x, base\]}<br/>{"$log": {"x": x, "base": base}}                          | element-wise for tensor types |
| map       | x: any[], f: logic<br/>x: tensor, f: logic                                                                                      | y: any[]<br/>y: tensor                   | {"$map": \[x, f]}<br/>{"$map": {"x": x, "f": f}}                                    |                               |
| reduce    | x: any[], f: logic, start: any<br/>x: tensor, f: logic, start: float32                                                          | y: any<br/>y: tensor                     | {"$reduce": \[x, f, start]}<br/>{"$map": {"x": x, "f": f, "start": start}}          |                               |
| filter    | x: any[], f: logic                                                                                                              | y: any[]                                 |                                                                                     |                               |
| sort      | x: scalar[]<br/>x: any[], f: logic                                                                                              | y: scalar[]<br/>y: any[]                 |                                                                                     |                               |
| floor     | x: numerical                                                                                                                    | y: numerical                             |                                                                                     |                               |
| in        | x: any, array: any[]                                                                                                            | y: bool                                  |                                                                                     |                               |
| append    | x: any, array: any[]                                                                                                            | y: any[]                                 |                                                                                     |                               |
| index     | x: any, array: any[]<br/>x: binary, array: binary                                                                               | y: int64                                 |                                                                                     |                               |
| time      |                                                                                                                                 | y: int64                                 |                                                                                     | unix timestamp milliseconds   |
| hour      | time: int64, tz: int64                                                                                                          | y: float64                               |                                                                                     |                               |
| dayOfWeek | time: int64, tz: int64                                                                                                          | y: uint64                                |                                                                                     |                               |
| day       | time: int64, tz: int64                                                                                                          | y: float64                               |                                                                                     |                               |
| month     | time: int64, tz: int64                                                                                                          | y: float64                               |                                                                                     |                               |
| year      | time: int64, tz: int64                                                                                                          | y: float64                               |                                                                                     |                               |
| north     | time: int64, tz: int64                                                                                                          | y: vector                                |                                                                                     |                               |
| south     | time: int64, tz: int64                                                                                                          | y: vector                                |                                                                                     |                               |
| west      | time: int64, tz: int64                                                                                                          | y: vector                                |                                                                                     |                               |
| east      | time: int64, tz: int64                                                                                                          | y: vector                                |                                                                                     |                               |
| sun       | time: int64, tz: int64                                                                                                          | y: vector                                |                                                                                     |                               |
| moon      | time: int64, tz: int64                                                                                                          | y: vector                                |                                                                                     |                               |
| mod       | x1: numerical, x2: numerical                                                                                                    | y: float64                               |                                                                                     |                               |
| zero      | type: string<br/>type: string, size: uint64<br/>type: string, size: ivector3                                                    | y: matrix<br/>y: scalar[]<br/>y: tensor3 | {$zero: type}<br/>{$zero: {"type": type, "size": size}}                             |                               |
| ones      | type: string<br/>type: string, size: uint64<br/>type: string, size: ivector3                                                    | y: matrix<br/>y: scalar[]<br/>y: tensor3 | {$ones: type}<br/>{$zero: {"type": type, "size": size}}                             |                               |
| getSlice  | x: any[], from: int64, to: int64<br/>x: tensor3, from: ivector3, to: ivector3                                                   | y: any[]<br/>y: tensor3                  |                                                                                     |                               |
| setSlice  | x: any[], from: int64, to: int64, value: any[]<br/>x: tensor3, from: ivector3, to: ivector3, value: tensor3                     | y: any[]<br/>y: tensor3                  |                                                                                     |                               |
| getProp   | x: props, key: string, value: any                                                                                               | y: any                                   |                                                                                     |                               |
| setProp   | x: props, key: string, value: any                                                                                               | y: props                                 |                                                                                     |                               |
| get       | index: int64, array: any[], value: any<br/>index: int64, array: bytes, value: byte<br/>index: int64, array: string, value: char | y: any<br/>y: byte<br/>y: char           |                                                                                     |                               |
| set       | index: int64, array: any[], value: any<br/>index: int64, array: bytes, value: byte<br/>index: int64, array: string, value: char | y: any[]<br/>y: bytes<br/>y: string      |                                                                                     |                               |
| ltrim     | x: string                                                                                                                       | y: string                                |                                                                                     |                               |
| rtrim     | x: string                                                                                                                       | y: string                                |                                                                                     |                               |
| max       | x: numerical[]<br/>x: tensor                                                                                                    | y: numerical                             |                                                                                     |                               |
| min       | x: numerical[]<br/>x: tensor                                                                                                    | y: numerical                             |                                                                                     |                               |
| average   | x: numerical[]<br/>x: tensor                                                                                                    | y: numerical                             |                                                                                     |                               |
| std       | x: numerical[]<br/>x: tensor                                                                                                    | y: numerical                             |                                                                                     |                               |
| break     | status: string                                                                                                                  | -                                        |                                                                                     | interrupts message processing |

- [ ] to be continued

## Example
### Default object
The default object. 
* have not any access control
* occupied one blockcell
* have not funds
* any next modifyProps message will just reflect into object's props
* more strict logic can be applied by modifying corresponding handler props
```json
{
  "blockcells": [[0, 0, 0]],
  "bcOrigin": [0, 0, 0],
  "balance": 0,
  "msgValidation": {
    "name": "Default message validation logic",
    "outputs": {
      "isValid": {
        "name": "is incoming message valid",
        "valueType": "bool",
        "value": true
      }
    }
  },
  "shootHandler": {"$get_by_id": <uint256>},
  "modifyObjectHandler": {
    "name": "Default modify logic",
    "inputs": {
      "object": {
        "name": "Object properties",
        "valueType": "prop[]",
        "defaultValue": null
      },
      "message": {
        "name": "Message properties",
        "valueType": "prop[]",
        "defaultValue": null
      }
    },
    "outputs": {
      "props": {
        "name": "props to update for the object's state",
        "valueType": "prop[]",
        "value": "$message"
      },
      "events": {
        "name": "events to fire when object's state updated",
        "valueType": "props[]",
        "value": [{"$concat": [{"objectId": "$object.id"}, "$message"]}]
      },
      "messages": {
        "name": "messages to send when object's state updated",
        "valueType": "props[]",
        "value": null
      }
    }
  }
}
```

### Farming object







------
ray(start, direction, length)
ssector(start, direction, angle)

avg?
cmp?
covariance?



