package core

import (
	"log"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/dsl/errors"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/dsl/interfaces"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/dsl/typing"
	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"
)

type Closure struct {
	functor interface {
		interfaces.Functor
	}
	arg interface {
		interfaces.Getter
		interfaces.Keyable
		interfaces.Setter
		interfaces.Sized
		interfaces.Evaluable
		interfaces.Typable
	}
}

func BuildClosureByRef(ref *Ref, payload *aproto.Payload, ctx interfaces.Context) (clsr *Closure, err error) {
	if fn, ok := ref.Val(ctx).(interfaces.Functor); ok {
		clsr, err = BuildClosure(fn, payload, ctx)
	} else {
		err = errors.New("ONLY FUNCTERS ARE ALLOWED FOR CLOSURE")
	}
	return
}

func BuildClosure(fn interfaces.Functor, payload *aproto.Payload, ctx interfaces.Context) (clsr *Closure, err error) {
	var arg interfaces.Unit
	if arg, err = Build(payload, ctx); err != nil {
		return
	}

	clsr = &Closure{
		functor: fn,
	}

	switch fn := fn.(type) {
	case interfaces.Functor:
		var (
			spec    = fn.InputSpec()
			props   *Props
			isProps bool
		)

		if props, isProps = arg.(*Props); !isProps {
			if list, isList := arg.(List); isList && len(list) <= spec.Size() {
				props = NewProps(nil)
				keys := spec.Keys()
				for i, unit := range list {
					props.Set(keys[i], unit)
				}
			} else {
				err = errors.INVALID_SYNTAX_ERR
			}
		}

		if err == nil {
			// Squash
			if closureFn, ok := fn.(*Closure); ok {
				clsr.functor = closureFn.functor
				for _, key := range closureFn.arg.Keys() {
					props.Set(key, closureFn.arg.Get(key))
				}
			}

			clsr.arg = props
			for _, key := range spec.Keys() {
				spec := spec.Get(key)
				if prop := props.Get(key); prop == nil {
					err = errors.EXTRA_ARG_ERR
				} else {
					err = typing.Compare(spec.Type(ctx), prop.Type(ctx))
				}
				if err != nil {
					break
				}
			}
		}
	default:
		err = errors.New("BAD FUNCTOR TYPE")
	}

	if err != nil {
		clsr = nil
	}

	return
}

func (clsr *Closure) InputSpec() interface {
	interfaces.Getter
	interfaces.Keyable
	interfaces.Sized
	interfaces.Typable
} {
	if fn, ok := clsr.functor.(interfaces.Functor); ok {
		if clsr.arg.Size() == 0 {
			return fn.InputSpec()
		}

		spec := NewProps(nil)
		for _, key := range fn.InputSpec().Keys() {
			if clsr.arg.Get(key) == nil {
				spec.Set(key, fn.InputSpec().Get(key))
			}
		}
		return spec
	} else {
		log.Fatalln(errors.NOT_IMPLEMENTED_ERR)
	}
	return nil
}

// Impl interfaces.Evaluable
func (clsr *Closure) Eval(ctx interfaces.Context) (interfaces.Unit, error) {

	unit, err := clsr.arg.Eval(ctx)

	if err == nil {
		switch arg := unit.(type) {
		case interface {
			interfaces.Getter
			interfaces.Setter
		}:
			ctx.Push(arg)
			unit, err = clsr.functor.Eval(ctx)
			ctx.Pop()
		default:
			err = errors.New("BAD INPUT")
		}
	}

	return unit, err
}

// Impl interfaces.Typable
func (clsr *Closure) Type(ctx interfaces.Context) (t typing.Type) {
	ctx.Push(clsr.arg)
	t = clsr.functor.Type(ctx)
	ctx.Pop()
	return
}
