package core

import (
	"fmt"
	"strings"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/dsl/errors"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/dsl/interfaces"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/dsl/typing"
	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"
)

const (
	PropSep = '.'
)

type Props struct {
	props map[string]*Prop
	order []string
}

func NewProps(items []interfaces.Unit) (ps *Props) {
	ps = &Props{
		props: make(map[string]*Prop),
		order: make([]string, 0),
	}
	for _, item := range items {
		var key string
		if named, ok := item.(interfaces.Named); ok {
			key = named.Name()
		}
		if key == "" {
			key = fmt.Sprintf("%d", ps.Size())
		}
		ps.Set(key, item)
	}
	return
}

func BuildProps(props []*aproto.Prop, ctx interfaces.Context) (ps *Props, err error) {
	ps = NewProps(nil)

	for _, prop := range props {
		var p *Prop
		if p, err = BuildProp(prop, ctx); err == nil {
			ps.Set(prop.GetName(), p)
		} else {
			break
		}
	}

	if err != nil {
		ps = nil
	}

	return
}

// Impl interfaces.Evaluable
func (ps *Props) Eval(ctx interfaces.Context) (unit interfaces.Unit, err error) {
	evaled := &Props{
		props: make(map[string]*Prop),
		order: make([]string, 0, ps.Size()),
	}
	for _, key := range ps.order {
		var val interfaces.Unit
		if val, err = ps.props[key].Eval(ctx); err == nil {
			evaled.Set(key, val)
		}
		if err != nil {
			break
		}
	}

	if err == nil {
		unit = evaled
	}

	return
}

// Impl interfaces.Payloadable
func (props *Props) Payload(ctx interfaces.Context) (payload *aproto.Payload, err error) {
	propsPayload := make([]*aproto.Prop, 0)
	for _, key := range props.Keys() {
		prop := props.props[key]

		var (
			value        *aproto.Payload
			specDefValue *aproto.Payload
		)

		payloadable, ok := prop.Value.(interfaces.Payloadable)
		if !ok {
			val, err = prop.Value.Eval(ctx)
			payloadable, ok = val.(interfaces.Payloadable)
		}

		if ok {
			payload, err = payloadable.Payload(ctx)
			propsPayload = append(propsPayload, &aproto.Prop{
				Name:  key,
				Value: payload,
			})

		} else if err == nil {
			err = errors.NOT_FULLY_EVALED_ERR
		}

		if err != nil {
			break
		}
	}
	if err == nil {
		payload = &aproto.Payload{
			Value: &aproto.Payload_Props{
				Props: &aproto.PropList{
					Props: propsPayload,
				},
			},
		}
	}

	return
}

// Impl interfaces.Typable
func (ps *Props) Type(ctx interfaces.Context) typing.Type {
	props := make(typing.PropsType)
	for _, key := range ps.order {
		props[key] = ps.props[key].Type(ctx)
	}
	return props
}

// Impl interfaces.Getter
func (ps *Props) Get(key string) (val interfaces.Unit) {
	if key == "" {
		key = ArgPrefix
	}

	if val = ps.props[key]; val == nil {
		if splitInd := strings.Index(key, string(PropSep)); splitInd != -1 {
			if prop := ps.props[key[:splitInd]]; prop != nil {
				if getter, ok := prop.Value.(interfaces.Getter); ok {
					val = getter.Get(key[splitInd+1:])
				}
			}
		}
	}

	return
}

func (ps *Props) Keys() []string {
	return ps.order
}

// Impl interfaces.Setter
func (ps *Props) Set(key string, val interfaces.Unit) {
	var (
		p  *Prop
		ok bool
	)

	if p, ok = val.(*Prop); !ok {
		p = &Prop{
			Name:  key,
			Value: val,
		}
	}

	if _, ok := ps.props[key]; !ok {
		ps.order = append(ps.order, key)
	}
	ps.props[key] = p
}

func (ps *Props) Size() int {
	return len(ps.props)
}

func (ps *Props) Copy() (c *Props) {
	c = &Props{
		props: make(map[string]*Prop),
		order: make([]string, 0, ps.Size()),
	}
	for _, key := range ps.order {
		c.props[key] = ps.props[key]
		c.order = append(c.order, key)
	}
	return
}
