package core

import (
	"log"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/dsl/errors"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/dsl/interfaces"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/dsl/typing"
	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"
)

type Value struct {
	Data *aproto.Payload
}

// Impl interfaces.Evaluable
func (val *Value) Eval(ctx interfaces.Context) (interfaces.Unit, error) {
	return val, nil
}

// Impl interfaces.Payloadable
func (val *Value) Payload(ctx interfaces.Context) (*aproto.Payload, error) {
	return val.Data, nil
}

// Impl interfaces.Typable
func (val *Value) Type(ctx interfaces.Context) (t typing.Type) {

	switch val := val.Data.GetValue().(type) {
	case *aproto.Payload_Functor:
		log.Fatalln(errors.NOT_IMPLEMENTED_ERR)
	case *aproto.Payload_Props:
		props := typing.PropsType{}
		for _, prop := range val.Props.GetProps() {
			props[prop.GetName()] = (&Value{prop.GetValue()}).Type(ctx)
		}
		t = props
	case *aproto.Payload_ListBool:
		t = &typing.ListType{NestedType: typing.TypeBool, Shape: val.ListBool.Shape}
	case *aproto.Payload_ListBytes:
		t = &typing.ListType{NestedType: typing.TypeBytes, Shape: val.ListBytes.Shape}
	case *aproto.Payload_ListInt:
		t = &typing.ListType{NestedType: typing.TypeInt, Shape: val.ListInt.Shape}
	case *aproto.Payload_ListUint:
		t = &typing.ListType{NestedType: typing.TypeUint, Shape: val.ListUint.Shape}
	case *aproto.Payload_ListFloat:
		t = &typing.ListType{NestedType: typing.TypeFloat, Shape: val.ListFloat.Shape}
	case *aproto.Payload_ListString:
		t = &typing.ListType{NestedType: typing.TypeString, Shape: val.ListString.Shape}
	case *aproto.Payload_Singular:
		switch val.Singular.GetValue().(type) {
		case *aproto.Singular_BoolValue:
			t = typing.TypeBool
		case *aproto.Singular_BytesValue:
			t = typing.TypeBytes
		case *aproto.Singular_IntValue:
			t = typing.TypeInt
		case *aproto.Singular_UintValue:
			t = typing.TypeUint
		case *aproto.Singular_FloatValue:
			t = typing.TypeFloat
		case *aproto.Singular_StringValue:
			t = typing.TypeString
		}
	}

	return
}

type List []interfaces.Unit

func BuildList(payload *aproto.Payload, ctx interfaces.Context) (unit interfaces.Unit, err error) {
	list := make(List, 0)
	var childUnit interfaces.Unit
	for _, childPayload := range payload.GetListPayload().GetValue() {
		childUnit, err = Build(childPayload, ctx)
		if err == nil {
			list = append(list, childUnit)
		} else {
			break
		}
	}

	if err == nil {
		unit = list
	}

	return
}

// Impl interfaces.Evaluable
func (list List) Eval(ctx interfaces.Context) (interfaces.Unit, error) {
	var err error

	unitList := make(List, len(list))

	for i, unit := range list {
		if unitList[i], err = unit.Eval(ctx); err != nil {
			break
		}
	}

	return unitList, err
}

// Impl interfaces.Typable
func (list List) Type(ctx interfaces.Context) typing.Type {
	log.Panicln(errors.NOT_IMPLEMENTED_ERR)
	return nil
}
