package typing

import (
	"reflect"
	"strconv"
	"strings"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/dsl/errors"
)

func Parse(nota string) (t Type, err error) {
	if len(nota) == 0 {
		err = errors.INCORRECT_TYPE_NOTATION_ERR
		return
	}

	if nota[0] == '[' {
		if endInd := strings.IndexRune(nota, ']'); endInd != -1 {
			size := -1
			if endInd > 1 {
				size, err = strconv.Atoi(nota[:endInd])
			}
			if err == nil {
				var nestedType Type
				if nestedType, err = Parse(nota[endInd+1:]); err == nil {
					listType := &ListType{
						Shape: []int64{int64(size)},
					}
					if nestedListType, isList := nestedType.(*ListType); isList {
						listType.Shape = append(listType.Shape, nestedListType.Shape...)
						listType.NestedType = nestedListType.NestedType
					} else {
						listType.NestedType = nestedType
					}
					t = listType
				}
			}
		}
	} else if t = builtInTypes.byNotation[nota]; t == nil {
		err = errors.INCORRECT_TYPE_NOTATION_ERR
	}

	return
}

func Compare(dst Type, src Type) (err error) {
	if !dst.Validate(src) {
		if src == TypeBad {
			err = errors.BAD_TYPE_ERR
		} else {
			err = errors.INVALID_TYPE_ERR
		}
	}
	return
}

type Type interface {
	GetNotation() string
	Validate(inType Type) (ok bool)
}

type BasicType struct {
	notation   string
	nativeType reflect.Kind
	protoType  reflect.Kind
}

func newBasicType(notation string, nativeType interface{}, protoType interface{}) *BasicType {
	return &BasicType{
		notation:   notation,
		nativeType: reflect.TypeOf(nativeType).Kind(),
		protoType:  reflect.TypeOf(protoType).Kind(),
	}
}

func (t *BasicType) GetNotation() string {
	return t.notation
}

func (t *BasicType) Validate(inType Type) bool {
	basicInType, ok := inType.(*BasicType)
	if !ok {
		return ok
	}

	return t.nativeType == basicInType.nativeType
}

type ListType struct {
	notation   string
	NestedType Type
	protoType  reflect.Kind
	Shape      []int64
}

func (t *ListType) GetNotation() string {
	return t.notation
}

func (t *ListType) Validate(inType Type) (ok bool) {
	var listInType *ListType
	if listInType, ok = inType.(*ListType); !ok {
		return
	}

	if ok = len(t.Shape) == len(listInType.Shape); ok {
		for i := range t.Shape {
			if ok = t.Shape[i] == listInType.Shape[i]; !ok {
				break
			}
		}
	}

	if ok {
		ok = t.NestedType.Validate(listInType.NestedType)
	}

	return
}

type SetType struct {
	notation string
	Types    []Type
}

func (t *SetType) GetNotation() string {
	return t.notation
}

func (t *SetType) Validate(inType Type) (ok bool) {
	for _, nestedType := range t.Types {
		if ok = nestedType.Validate(inType); ok {
			break
		}
	}
	return ok
}

type PropsType map[string]Type

func (t PropsType) GetNotation() string {
	return "props"
}

func (t PropsType) Validate(inType Type) (ok bool) {
	var mapInType PropsType
	if mapInType, ok = inType.(PropsType); !ok {
		return
	}
	for key, valType := range t {
		if inType, ok = mapInType[key]; ok {
			ok = valType.Validate(inType)
		}
		if !ok {
			break
		}
	}

	return
}

type CustomType struct {
	FnNotation func() string
	FnValidate func(inType Type) (ok bool)
}

func (t *CustomType) GetNotation() string {
	return t.FnNotation()
}

func (t *CustomType) Validate(inType Type) (ok bool) {
	return t.FnValidate(inType)
}
