package grpc

import (
	"context"
	"fmt"
	"net"

	"go.uber.org/zap"
	"google.golang.org/grpc"
	"google.golang.org/grpc/metadata"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/config"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/exec"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang/proto"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/subscribe"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/utils"
	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"
)

const clientIDHeaderField = "client_id"

var logger = utils.GetLogger("GRPC_SERVER")

type L3ServiceServer struct {
	aproto.UnimplementedL3ServiceServer

	ctx            *context.Context
	subscribeMngr  *subscribe.Manager
	execController *exec.ExecController
}

func (s *L3ServiceServer) Stream(stream aproto.L3Service_StreamServer) (err error) {
	headers, ok := metadata.FromIncomingContext(stream.Context())
	if !ok {
		err = fmt.Errorf("HEADERS NOT PROVIDED")
		return
	}

	var clientID *proto.ID
	if keys := headers.Get(clientIDHeaderField); len(keys) != 0 {
		clientID, err = proto.MakeIDFromBase64(keys[0])
	} else {
		err = fmt.Errorf("%s not provided", clientIDHeaderField)
		return
	}

	if err != nil {
		return
	}

	sbr := subscribe.NewSubscriber(clientID)

	s.subscribeMngr.RegisterSubscriber(sbr)
	defer func() {
		s.subscribeMngr.UnregisterSubscriber(sbr)
	}()

	// Sender
	go func() {
		for {
			response, ok := sbr.RecvResponse()
			if !ok {
				logger.Debug("Subscriber closed", zap.String("client_key", clientID.Base64()))
				break
			}
			if err = stream.Send(response); err != nil {
				logger.Debug("Fail to send response", zap.Error(err))
				break
			}
			logger.Debug("Response sended to GRPC client", zap.String("client_key", clientID.Base64()))
		}
	}()

	// Receiver
	var txn *aproto.TxnRequest
	for {
		if txn, err = stream.Recv(); err != nil {
			logger.Debug("Fail to recv msg", zap.Error(err))
			break
		}
		if err = s.execController.PushTxn(txn); err != nil {
			logger.Debug("Fail to add txn to queue", zap.Error(err))
			break
		}
	}

	return
}

type Server struct {
	server *grpc.Server
}

func NewServer(ctx context.Context) *Server {
	return &Server{
		server: grpc.NewServer(),
	}
}

func (s *Server) Start(subscribeMngr *subscribe.Manager, execController *exec.ExecController) (err error) {
	tcp, err := net.Listen("tcp", fmt.Sprintf("0.0.0.0:%d", config.GRPC_PORT))
	if err != nil {
		return
	}
	defer tcp.Close()

	aproto.RegisterL3ServiceServer(s.server, &L3ServiceServer{
		subscribeMngr:  subscribeMngr,
		execController: execController,
	})

	if err = s.server.Serve(tcp); err != nil {
		logger.Error("Server error", zap.Error(err))
	}

	return
}

func (s *Server) GracefulStop() {
	s.server.GracefulStop()
}

func (s *Server) Stop() {
	s.server.Stop()
}
