package ws

import (
	"context"
	"fmt"
	"net"
	"net/http"

	"github.com/gorilla/mux"
	"github.com/gorilla/websocket"
	"go.uber.org/zap"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/config"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/exec"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang/proto"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/subscribe"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/utils"
	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"
)

var logger = utils.GetLogger("WS_SERVER")

var upgrader = websocket.Upgrader{}

func wsHandlerWrapper(subscribeMngr *subscribe.Manager, execController *exec.ExecController) http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		clientID, err := proto.MakeIDFromBase64(mux.Vars(r)["client_id"])
		if err != nil {
			return
		}
		// TODO: Auth

		conn, err := upgrader.Upgrade(w, r, nil)
		if err != nil {
			logger.Error("Fail to open WS connection", zap.Error(err))
			return
		}
		defer conn.Close()

		sbr := subscribe.NewSubscriber(clientID)

		subscribeMngr.RegisterSubscriber(sbr)
		defer func() {
			subscribeMngr.UnregisterSubscriber(sbr)
		}()

		// Sender
		go func() {
			defer conn.Close()

			for {
				response, ok := sbr.RecvResponse()
				if !ok {
					logger.Debug("Subscriber closed", zap.String("client_key", clientID.Base64()))
					break
				}

				bts, err := proto.Marshal(response)
				if err == nil {
					err = conn.WriteMessage(websocket.BinaryMessage, bts)
				}

				if err != nil {
					logger.Error("Fail to send response", zap.Error(err))
					break
				}
			}
		}()

		// Receiver
		for {
			_, bts, err := conn.ReadMessage()
			if err != nil {
				break
			}

			var txn aproto.TxnRequest
			if err = proto.Unmarshal(bts, &txn); err != nil {
				break
			}

			if err = execController.PushTxn(&txn); err != nil {
				break
			}
		}
	}
}

type Server struct {
	tcp net.Listener
}

func NewServer(ctx context.Context) *Server {
	return &Server{}
}

func (s *Server) Start(subscribeMngr *subscribe.Manager, execController *exec.ExecController) (err error) {
	s.tcp, err = net.Listen("tcp", fmt.Sprintf("0.0.0.0:%d", config.WS_PORT))
	if err != nil {
		return
	}
	defer s.tcp.Close()

	r := mux.NewRouter()
	r.HandleFunc("/ws/{client_key}", wsHandlerWrapper(subscribeMngr, execController))

	if err = http.Serve(s.tcp, nil); err != nil {
		logger.Error("Server error", zap.Error(err))
	}

	return
}

func (s *Server) Stop() {
	if s.tcp != nil {
		s.tcp.Close()
	}
}
