package exec

import (
	"sync"
	"time"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/store"
	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"
)

const (
	maxBatchSize = 10
)

type Batch struct {
	time time.Duration

	txns []*aproto.TxnRequest
	msgs []*aproto.Msg

	externalMsgCount int
	internalMsgCount int
}

func NewBatch() *Batch {
	return &Batch{}
}

func (b *Batch) AddTxn(txn *aproto.TxnRequest) {
	b.externalMsgCount += len(txn.Messages)
	b.txns = append(b.txns, txn)
}

func (b *Batch) AddMessages(msgs []*aproto.Msg) {
	b.internalMsgCount += len(msgs)
	b.msgs = append(b.msgs, msgs...)
}

func (b *Batch) Size() int {
	return b.externalMsgCount + b.internalMsgCount
}

func (b *Batch) Full() bool {
	return b.Size() >= maxBatchSize
}

func (b *Batch) ToRollupCycle() *aproto.Rollup_Cycle {
	return &aproto.Rollup_Cycle{
		Time:         uint64(b.time),
		Txns:         b.txns,
		InternalMsgs: uint64(b.internalMsgCount),
	}
}

func (b *Batch) Process(exec *ExecController) (err error) {

	execUnits := make([]*ExecUnit, 0, b.Size())

	for _, txn := range b.txns {
		for _, msg := range txn.Messages {
			execUnits = append(execUnits, NewExecUnit(exec, msg, true))
		}
	}

	for _, msg := range b.msgs {
		execUnits = append(execUnits, NewExecUnit(exec, msg, false))
	}

	for work := true; work; {
		registryBase := store.NewRegistryBase()
		wg := sync.WaitGroup{}

		for i, unit := range execUnits {
			if !unit.IsDone() {
				reg := registryBase.NewRegistry(uint(i))
				if unit.Init(exec.store, reg) == nil {
					wg.Add(1)
					go func(c *ExecUnit) {
						defer wg.Done()
						c.Exec()
					}(unit)
				}
			}
		}
		wg.Wait()

		work = false
		for _, unit := range execUnits {
			if !unit.IsDone() {
				work = true
				break
			}
		}
	}

	for _, unit := range execUnits {
		unit.Release()
	}
	exec.store.Commit()

	err = exec.rollupProc.AddCycle(b.ToRollupCycle())

	return
}
