package exec

import (
	"net/http"

	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"
	"go.uber.org/zap"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/dapp"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/errors"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang/proto"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/store"
)

type ExecUnit struct {
	exec *ExecController

	// INPUT
	callerID   *proto.ID
	msg        *aproto.Msg
	objectID   *proto.ID
	isExternal bool

	// LOADED SUBJECTS
	payer    *store.Object
	msgProps *proto.Props
	object   *store.Object

	// RUNTIME
	runtime *lang.Runtime

	// STATE
	executed bool
	err      error
}

func NewExecUnit(exec *ExecController, msg *aproto.Msg, isExternal bool) (unit *ExecUnit) {
	unit = &ExecUnit{
		exec:       exec,
		callerID:   proto.NewID(msg.Sender),
		msg:        msg,
		isExternal: isExternal,
		objectID:   proto.NewID(msg.Target),
	}

	return
}

func (unit *ExecUnit) IsDone() bool {
	return unit.executed || unit.err != nil
}

func (unit *ExecUnit) Init(st *store.Store, registry *store.Registry) (err error) {
	defer func() {
		unit.err = err
	}()

	if unit.msgProps == nil {
		unit.msgProps = proto.NewProps(unit.msg.Props)
		if !unit.msgProps.Sorted() {
			err = errors.ErrPropsNotSorted
			return
		}
	}

	wrappedSt, err := st.ExecWrapper(registry)
	if err != nil {
		return
	}

	unit.payer, err = st.Object().Get(unit.callerID)
	if err != nil {
		return
	}
	if err = unit.payer.LoadBalance(); err != nil {
		return
	}

	unit.object, err = wrappedSt.Object().Get(unit.objectID)
	if err != nil {
		return
	}
	if unit.callerID.Base64() == unit.objectID.Base64() {
		unit.object.Balance = unit.payer.Balance
	} else if err = unit.object.LoadBalance(); err != nil {
		return
	}

	unit.runtime = lang.NewRuntime(unit.payer.Balance, unit.msgProps, unit.object, wrappedSt, unit.exec.lib)
	return
}

func (unit *ExecUnit) Exec() (err error) {
	defer func() {
		unit.executed = err == nil || err != errors.ErrInterrupt
		if unit.executed {
			unit.err = err
		}
	}()

	var handler *aproto.Prop
	handler, err = unit.runtime.Object().Get(unit.msg.Type)

	if err == nil && handler == nil {
		err = errors.ErrHandlerNotFound
	}

	if err != nil {
		return
	}

	_, err = lang.EvalPayload(handler.GetValue(), unit.runtime)

	return
}

func (unit *ExecUnit) Release() {
	if !unit.IsDone() {
		return
	}

	unit.payer.WriteBalance()
	if unit.err == nil {
		unit.runtime.Object().WithdrawHolded()
		unit.runtime.Object().WriteBalance()
		unit.runtime.Store().Write()

		unit.exec.PushMessages(unit.runtime.Messages())
		unit.exec.subscribeMngr.BroadcastMessages(unit.runtime.Events())
	}

	if unit.isExternal {
		codeProp := &aproto.Prop{
			Name:  "code",
			Value: proto.UintPayload(http.StatusOK),
		}
		msgBytes, _ := proto.Marshal(unit.msg)
		msg := &aproto.Msg{
			Sender: dapp.DApp.ID().Bytes(),
			Target: unit.msg.Sender,
			Type:   "message_exec_status",
			Props: []*aproto.Prop{
				codeProp,
				{
					Name:  "msg_bytes",
					Value: proto.BytesPayload(msgBytes),
				},
			},
		}

		if unit.err != nil {
			codeProp.Value = proto.UintPayload(http.StatusBadRequest)
			msg.Props = append(msg.Props, &aproto.Prop{
				Name:  "error",
				Value: proto.StringPayload(unit.err.Error()),
			})
			{
				var errMsg string
				if e, ok := unit.err.(*errors.Error); ok {
					errMsg = e.Msg
				}
				logger.Error("MESSAGE EXECUTION ERR",
					zap.Uint64("msg_id", unit.msg.Id),
					zap.Error(unit.err),
					zap.String("err_msg", errMsg),
				)

			}
		}

		unit.exec.subscribeMngr.BroadcastMessage(msg)
	}
}
