package compiler

import (
	"errors"
	"strconv"

	"golang.org/x/exp/utf8string"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang/proto"
	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"
)

const (
	GramPropSet     = '='
	GramParamsBegin = '<'
	GramParamsEnd   = '>'
	GramCallBegin   = '('
	GramCallEnd     = ')'
	GramPropsBegin  = '{'
	GramPropsEnd    = '}'
	GramListBegin   = '['
	GramListEnd     = ']'
	GramString      = '"'
	GramNameDelim   = '_'
	GramNeg         = '-'
	GramDot         = '.'
	GramComma       = ','
)

var (
	ErrInvalidSyntax          = errors.New("INVALID SYNTAX")
	ErrInvalidNameSymbol      = errors.New("INVALID NAME SYMBOL")
	ErrInvalidNameStartSymbol = errors.New("INVALID NAME START SYMBOL")
	ErrEmptyName              = errors.New("EMPTY NAME")
	ErrInvalidCall            = errors.New("INVALID CALL")

	ErrUnkownSymbol   = errors.New("UNKOWN SYMBOL")
	ErrNotImplemented = errors.New("NOT IMPLEMENTED")
	ErrBlockNotFound  = errors.New("BLOCK NOT FOUND")
	ErrValueNotFound  = errors.New("VALUE NOT FOUND")
	ErrStringNotFound = errors.New("STRING NOT FOUND")
	ErrListNotFound   = errors.New("LIST NOT FOUND")
)

type Parser struct {
	err error
	str *utf8string.String

	begin int
	cur   int

	curRune string
}

func newParser(str string) *Parser {
	return &Parser{
		str:   utf8string.NewString(str),
		begin: 0,
		cur:   0,
	}
}

func (p *Parser) isEnd() bool {
	return p.str.RuneCount() <= p.cur
}

func (p *Parser) isAlpha() bool {
	return ('a' <= p.rune() && p.rune() <= 'z') || ('A' <= p.rune() && p.rune() <= 'Z')
}

func (p *Parser) isDigit() bool {
	return '0' <= p.rune() && p.rune() <= '9'
}

func (p *Parser) isSpace() bool {
	switch p.rune() {
	case ' ', '\t', '\n':
		return true
	default:
		return false
	}
}

func (p *Parser) work() bool {
	return !p.isEnd() && p.err == nil
}

func (p *Parser) rune() rune {
	if !p.isEnd() {
		p.curRune = string(p.str.At(p.cur))
		return p.str.At(p.cur)
	}
	return 0
}

func (p *Parser) next() rune {
	p.cur++
	return p.rune()
}

func (p *Parser) moveToNonSpace() bool {
	for p.work() {
		if p.isSpace() {
			p.next()
		} else {
			break
		}
	}
	return p.work()
}

func (p *Parser) parseValue() (payload *aproto.Payload) {
	p.moveToNonSpace()
	if p.isEnd() {
		return
	}

	switch p.rune() {
	case GramPropsBegin:
		payload = proto.PropsPayload(p.parseProps(true))
	case GramListBegin:
		payload = p.parseList()
	case GramString:
		payload = proto.StringPayload(p.parseString())
	default:
		if p.isDigit() || p.rune() == GramNeg {
			payload = p.parseNumber()
		} else if p.isAlpha() {
			payload = proto.CallPayload(p.parseCall())
		} else {
			p.err = ErrInvalidSyntax
		}
	}

	return
}

func (p *Parser) parseProps(sort bool) (props []*aproto.Prop) {
	done := false
	for p.work() && !done {
		if p.rune() == GramPropsEnd || p.rune() == GramCallEnd {
			done = true
		} else if p.isSpace() {
			p.moveToNonSpace()
		} else if p.rune() == GramPropsBegin || p.rune() == GramCallBegin || p.rune() == GramComma {
			p.next()
		} else {
			props = append(props, p.parseProp())
		}
	}
	p.next()

	if !done {
		p.err = ErrInvalidSyntax
	}

	if sort {
		props = proto.NewProps(props).Sort().Props()
	}

	return
}

func (p *Parser) parseProp() (prop *aproto.Prop) {
	prop = &aproto.Prop{}
	for p.work() {
		if p.isSpace() {
			p.moveToNonSpace()
		} else if p.rune() == GramPropSet {
			p.next()
			prop.Value = p.parseValue()
			break
		} else if prop.Name == "" {
			prop.Name = p.parseName()
			if prop.Name != "" && prop.Name[0] == GramNameDelim {
				prop.Spec = &aproto.Spec{
					Private: true,
				}
			}
		} else {
			p.err = ErrInvalidSyntax
		}
	}

	return
}

func (p *Parser) parseName() (name string) {
	if !p.moveToNonSpace() {
		return
	}

	for p.work() && !p.isSpace() {
		if p.isAlpha() || p.rune() == GramNameDelim {
			name += string(p.rune())
			p.next()
		} else if name == "" {
			p.err = ErrInvalidNameStartSymbol
		} else if p.isDigit() {
			name += string(p.rune())
			p.next()
		} else if p.rune() == GramCallBegin || p.rune() == GramParamsBegin {
			break
		} else {
			p.err = ErrInvalidNameSymbol
		}
	}

	if p.err == nil && name == "" {
		p.err = ErrEmptyName
	}

	return
}

func (p *Parser) parseString() (str string) {
	if p.rune() == GramString {
		p.next()
	}
	for p.work() && p.rune() != GramString {
		str += string(p.rune())
		p.next()
	}
	p.next()
	return
}

func (p *Parser) parseNumber() (payload *aproto.Payload) {
	if !p.moveToNonSpace() {
		return
	}

	var (
		isFloat bool
		isUint  bool
		isEnd   bool
		buf     = ""
	)

	for !isEnd && p.work() && !p.isSpace() {
		if p.isDigit() {
			buf += string(p.rune())
		} else {
			switch p.rune() {
			case GramNeg:
				if buf != "" {
					p.err = ErrInvalidSyntax
				}
				buf += string(p.rune())
			case GramDot:
				if isFloat {
					p.err = ErrInvalidSyntax
				}
				isFloat = true
				if buf == "" || buf == "-" {
					buf += "0"
				}
				buf += string(p.rune())
			case 'f':
				isFloat = true
				isEnd = true
			case 'u':
				isUint = true
				isEnd = true
			case GramComma:
				isEnd = true
			default:
				p.err = ErrInvalidSyntax
			}
		}
		p.next()
	}

	if p.err == nil {
		var f64 float64
		f64, p.err = strconv.ParseFloat(buf, 64)
		if p.err == nil {
			if isFloat {
				payload = proto.FloatPayload(float32(f64))
			} else if isUint {
				if f64 < 0 {
					p.err = ErrInvalidSyntax
				}
				payload = proto.UintPayload(uint64(f64))
			} else {
				payload = proto.IntPayload(int64(f64))
			}
		}
	}

	return
}

func (p *Parser) parseCall() (call *aproto.Call) {
	call = &aproto.Call{
		Name: p.parseName(),
	}

	if !p.moveToNonSpace() {
		p.err = ErrInvalidCall
		return
	}

	if p.rune() == GramParamsBegin {
		call.Params = p.parseParams()
	}

	if !p.moveToNonSpace() {
		p.err = ErrInvalidCall
		return
	}

	if p.rune() == GramCallBegin {
		call.Args = p.parseProps(false)
	}

	return
}

func (p *Parser) parseParams() (params []string) {
	var buf string
	for p.work() && p.rune() != GramParamsEnd {
		if p.rune() == GramComma {
			if buf == "" {
				p.err = ErrInvalidSyntax
				break
			}
			params = append(params, buf)
			buf = ""
		} else if p.rune() != GramParamsBegin && !p.isSpace() {
			buf += string(p.rune())
		}
		p.next()
	}
	if buf != "" {
		params = append(params, buf)
	}
	p.next()

	return
}

func (p *Parser) parseList() (payload *aproto.Payload) {
	list := &aproto.PayloadList{}
	for p.work() && p.rune() != GramListEnd {
		if p.isSpace() {
			p.moveToNonSpace()
		} else if p.rune() == GramListBegin {
			p.next()
		} else if p.rune() == GramComma {
			p.next()
		} else {
			list.Value = append(list.Value, p.parseValue())
		}
	}
	p.next()

	payload, ok := proto.ListPayloadToSomeList(list)
	if !ok {
		p.err = ErrInvalidSyntax
	}

	return
}
