package lang

import (
	"log"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/errors"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang/typing"
	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"
)

type Library map[string]*Func

func (lib Library) Get(key string) *Func {
	return lib[key]
}

func (lib Library) RegisterFn(fn *Func) *Func {
	err := fn.build()
	if err != nil {
		log.Fatalf("Fail to register function \"%s\" with error: %s\n", fn.Name, err.Error())
	}
	lib[fn.Name] = fn
	return fn
}

type Arg struct {
	Name         string
	Type         typing.Type
	TypeFn       func(typeParams map[string]typing.Type) typing.Type
	DefaultValue *aproto.Payload

	SpecDesc string
}

func (arg *Arg) getType(typeParams map[string]typing.Type) typing.Type {
	if arg.Type != nil {
		return arg.Type
	}

	if arg.TypeFn != nil {
		return arg.TypeFn(typeParams)
	}

	return typing.TypeAny
}

type Func struct {
	Name string

	TypeParams []string

	Args           []Arg
	argsByName     map[string]*Arg
	AllowExtraArgs bool

	ReturnType   typing.Type
	ReturnTypeFn func(typeParams map[string]typing.Type) typing.Type

	Fn func(args map[string]*aproto.Prop, rt *Runtime) (evaled *aproto.Payload, err error)
}

func (fn *Func) build() (err error) {
	if fn.ReturnType == nil && fn.ReturnTypeFn == nil {
		err = errors.ErrTypeNotSpecified
		return
	}
	fn.argsByName = make(map[string]*Arg)
	for _, arg := range fn.Args {
		if arg.Type == nil && arg.TypeFn == nil {
			err = errors.ErrTypeNotSpecified
			return
		}
		fn.argsByName[arg.Name] = &arg
	}
	return
}

func (fn *Func) prepareArgs(args []*aproto.Prop) (argMap map[string]*aproto.Prop, err error) {
	argMap = make(map[string]*aproto.Prop)
	argsLeft := len(fn.Args)

	hasNamed := false
	for i, arg := range args {
		if arg.Name == "" {
			if hasNamed {
				err = errors.ErrUnnamedArgs
				break
			} else {
				arg.Name = fn.Args[i].Name
			}
		} else {
			hasNamed = true
		}

		if _, ok := fn.argsByName[arg.Name]; !ok {
			if !fn.AllowExtraArgs {
				err = errors.ErrExtraArgs
				break
			}
		} else if _, ok := argMap[arg.Name]; !ok {
			argsLeft--
		}

		argMap[arg.Name] = arg
	}

	if argsLeft != 0 {
		err = errors.ErrNotEnoughArgs
	}

	return
}

func (fn *Func) Eval(args []*aproto.Prop, rt *Runtime) (evaled *aproto.Payload, err error) {
	var argMap map[string]*aproto.Prop
	argMap, err = fn.prepareArgs(args)
	if err == nil {
		evaled, err = fn.Fn(argMap, rt)
	}
	return
}

func (fn *Func) Type(args []*aproto.Prop, typeNotas []string, rt *Runtime) (t typing.Type, err error) {
	if len(typeNotas) < len(fn.TypeParams) {
		err = errors.ErrNotEnoughTypeParams
		return
	}

	typeParams := make(map[string]typing.Type)
	{
		var typeParam typing.Type
		for i, nota := range typeNotas {
			typeParam, err = typing.Parse(nota)
			if err != nil {
				break
			}
			typeParams[fn.TypeParams[i]] = typeParam
		}
		if err != nil {
			return
		}
	}

	{
		var argMap map[string]*aproto.Prop
		argMap, err = fn.prepareArgs(args)
		if err != nil {
			if err == errors.ErrNotEnoughArgs {
				err = nil
				t = typing.TypeFuncRef
			} else {
				return
			}
		}

		for name, arg := range argMap {
			var argType typing.Type
			argType, err = TypeOfProp(arg, rt)
			if err != nil {
				break
			}
			err = typing.Check(fn.argsByName[name].getType(typeParams), argType)
			if err != nil {
				break
			}
		}

		if err != nil {
			return
		}
	}

	if err == nil && t == nil {
		if fn.ReturnType != nil {
			t = fn.ReturnType
		} else if fn.ReturnTypeFn != nil {
			t = fn.ReturnTypeFn(typeParams)
		} else {
			t = typing.TypeAny
		}
	}

	return
}
