package lib

import (
	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/errors"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang/proto"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang/typing"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/store"
)

var protectedProps = map[string]struct{}{
	"id":         {},
	"balance":    {},
	"nonce":      {},
	"allowanced": {},
}

var (
	_ = Lib.RegisterFn(&lang.Func{
		Name:       "get",
		TypeParams: []string{"T"},
		Args: []lang.Arg{
			{
				Name: "name",
				Type: typing.TypeString,
			},
		},
		ReturnTypeFn: func(typeParams map[string]typing.Type) typing.Type {
			return typeParams["T"]
		},
		Fn: func(input map[string]*aproto.Prop, rt *lang.Runtime) (evaled *aproto.Payload, err error) {
			name := input["name"].GetValue().GetSingular().GetStringValue()
			prop, err := rt.Object().Get(name)
			if err == nil {
				evaled = prop.GetValue()
				if evaled == nil {
					err = errors.ErrPropNotFound
				}
			}
			return
		},
	})

	_ = Lib.RegisterFn(&lang.Func{
		Name: "set",
		Args: []lang.Arg{
			{
				Name: "name",
				Type: typing.TypeString,
			},
			{
				Name: "value",
				Type: typing.TypeAny,
			},
		},
		ReturnType: typing.TypeBool,
		Fn: func(args map[string]*aproto.Prop, rt *lang.Runtime) (evaled *aproto.Payload, err error) {
			obj := rt.Object()
			if obj == nil {
				err = errors.ErrObjectUndefined
				return
			}

			name := args["name"].GetValue().GetSingular().GetStringValue()
			value := args["value"].GetValue()

			if _, ok := protectedProps[name]; ok {
				err = errors.ErrProtectedProp
				return
			}

			err = obj.Set(&aproto.Prop{
				Name:  name,
				Value: value,
			})

			evaled = proto.BoolPayload(err == nil)

			return
		},
	})

	_ = Lib.RegisterFn(&lang.Func{
		Name:       "getObject",
		TypeParams: []string{"T"},
		Args: []lang.Arg{
			{
				Name: "id",
				Type: typing.TypeBytes,
			},
			{
				Name: "name",
				Type: typing.TypeString,
			},
		},
		ReturnTypeFn: func(typeParams map[string]typing.Type) typing.Type {
			return typeParams["T"]
		},
		Fn: func(args map[string]*aproto.Prop, rt *lang.Runtime) (evaled *aproto.Payload, err error) {
			id := proto.NewID(args["id"].GetValue().GetSingular().GetBytesValue())
			name := args["name"].GetValue().GetSingular().GetStringValue()

			var object *store.Object
			object, err = rt.Store().Object().GetForRead(id)
			if err == nil && object.IsEmpty() {
				err = errors.ErrObjectNotFound
			}

			if err != nil {
				return
			}

			var prop *aproto.Prop
			prop, err = object.Get(name)
			if err == nil {
				evaled = prop.GetValue()
			}

			return
		},
	})

	_ = Lib.RegisterFn(&lang.Func{
		Name: "addObject",
		Args: []lang.Arg{
			{
				Name: "props",
				Type: typing.TypeProps,
			},
			{
				Name: "balance",
				Type: typing.TypeBytes,
			},
		},
		ReturnType: typing.TypeBool,
		Fn: func(args map[string]*aproto.Prop, rt *lang.Runtime) (evaled *aproto.Payload, err error) {

			var (
				id     *proto.ID
				object *store.Object
			)

			for object.IsEmpty() {
				object = rt.Store().Object().New()
			}

			var balance *proto.BigUint
			balance, err = proto.BigUintFromBytes(args["balance"].GetValue().GetSingular().GetBytesValue())
			if err != nil {
				return
			}

			{
				var (
					parentBalanceProp *aproto.Prop
					parentBalance     *proto.BigUint
				)
				parentBalanceProp, err = rt.Object().Get("balance")
				if err != nil {
					return
				}
				parentBalance, err = proto.BigUintFromBytes(parentBalanceProp.GetValue().GetSingular().GetBytesValue())
				if err != nil {
					return
				}

				if parentBalance.Lt(balance) {
					err = errors.ErrNotEnoughBalance
				}

				parentBalance.Sub(balance)
				parentBalanceProp.Value = parentBalance.Payload()
				rt.Object().Set(parentBalanceProp)
			}

			{
				object.Set(&aproto.Prop{
					Name:  "id",
					Value: id.Payload(),
				})
				object.Set(&aproto.Prop{
					Name:  "balance",
					Value: balance.Payload(),
				})
				object.Set(&aproto.Prop{
					Name:  "nonce",
					Value: proto.UintPayload(0),
				})
				object.Set(&aproto.Prop{
					Name:  "allowanced",
					Value: proto.PropsPayload(nil),
				})
			}

			props := args["props"].GetValue().GetProps().GetProps()
			for _, prop := range props {
				if _, ok := protectedProps[prop.Name]; ok {
					err = errors.ErrProtectedProp
					break
				}
				object.Set(prop)
			}

			if err == nil {
				evaled = id.Payload()
			}

			return
		},
	})
)
