package lang

import (
	"fmt"
	"sync"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/errors"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang/proto"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/store"
	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"
)

type Node interface{}

type gasBalance interface {
	Withdraw(*proto.BigUint) error
}

type Runtime struct {
	buildIn Library

	store *store.Store

	// input
	gas    gasBalance
	msg    *proto.Props
	object *store.Object

	// returns
	messages []*aproto.Msg
	events   []*aproto.Msg

	// tracing
	err         error
	activeNodes map[Node]struct{}
	callStack   []Node

	// sync
	mutex       sync.RWMutex
	interrupted bool
}

func NewRuntime(gas gasBalance, msg *proto.Props, object *store.Object, st *store.Store, buildIn Library) (rt *Runtime) {
	rt = &Runtime{
		buildIn: buildIn,

		store:  st,
		gas:    gas,
		msg:    msg,
		object: object,

		messages: make([]*aproto.Msg, 0),

		activeNodes: make(map[Node]struct{}),
		callStack:   make([]Node, 0),
		mutex:       sync.RWMutex{},
	}
	return
}

func (rt *Runtime) BuildIn() Library {
	return rt.buildIn
}

func (rt *Runtime) Store() *store.Store {
	return rt.store
}

func (rt *Runtime) Msg() *proto.Props {
	return rt.msg
}

func (rt *Runtime) Object() *store.Object {
	return rt.object
}

func (rt *Runtime) TypeCheck() (err error) {
	if rt == nil || rt.Object() == nil {
		return
	}

	rt.Object().Iterate(func(prop *aproto.Prop) bool {
		_, err = TypeOfProp(prop, rt)
		return err == nil
	})
	return
}

func (rt *Runtime) Eval() (evaled []*aproto.Prop, err error) {
	if rt == nil || rt.Object() == nil {
		return
	}

	rt.Object().Iterate(func(prop *aproto.Prop) bool {
		prop, err = EvalProp(prop, rt)
		if err != nil {
			return false
		}
		evaled = append(evaled, prop)
		return true
	})
	return
}

func (rt *Runtime) EvalByName(name string) (evaled *aproto.Prop, err error) {
	if rt == nil || rt.Object() == nil {
		return
	}

	evaled, err = rt.Object().Get(name)
	if err == nil {
		evaled, err = EvalProp(evaled, rt)
	}
	return
}

// RETURNS

func (rt *Runtime) RegisterMsg(msg *aproto.Msg) {
	rt.messages = append(rt.messages, msg)
}

func (rt *Runtime) Messages() []*aproto.Msg {
	return rt.messages
}

func (rt *Runtime) RegisterEvent(msg *aproto.Msg) {
	rt.events = append(rt.events, msg)
}

func (rt *Runtime) Events() []*aproto.Msg {
	return rt.events
}

// FLOW CONTROL

func (rt *Runtime) GasWithdraw(gas *proto.BigUint) (err error) {
	if gas != nil {
		err = rt.gas.Withdraw(gas)
	}
	return
}

func (rt *Runtime) Interrupt() {
	// rt.mutex.Lock()
	// defer rt.mutex.Unlock()
	rt.interrupted = true
}

func (rt *Runtime) Interrupted() bool {
	// rt.mutex.RLock()
	// defer rt.mutex.RUnlock()
	return rt.interrupted
}

func (rt *Runtime) EntryToNode(node Node) (err error) {
	rt.callStack = append(rt.callStack, node)

	if rt.Interrupted() {
		err = errors.ErrInterrupt
	} else if _, ok := rt.activeNodes[node]; !ok {
		rt.activeNodes[node] = struct{}{}
	} else {
		err = errors.ErrCycleDetected
	}

	return
}

func (rt *Runtime) ExitFromNode(node Node, err error) {
	delete(rt.activeNodes, node)

	if err == nil {
		n := len(rt.callStack) - 1
		rt.callStack = rt.callStack[:n]
	} else {
		rt.err = err
	}
}

func (rt *Runtime) StackTrace() (trace string) {
	for i, node := range rt.callStack {
		if i != 0 {
			trace += " -> "
		}

		switch val := node.(type) {
		case *aproto.Prop:
			trace += fmt.Sprintf("Prop[%s]", val.Name)
		case *aproto.Call:
			trace += fmt.Sprintf("Call[%p]", val)
		default:
			trace += "UNKNOWN"
		}
	}
	return
}
