package lang

import (
	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/errors"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang/typing"
)

func TypeOfPayload(payload *aproto.Payload, rt *Runtime) (t typing.Type, err error) {
	if payload == nil || rt == nil {
		return
	}

	switch val := payload.GetValue().(type) {
	case *aproto.Payload_Props:
		t, err = TypeOfProps(val.Props.GetProps(), rt)
	case *aproto.Payload_NoEvall:
		t, err = TypeOfPayload(val.NoEvall.GetPayload(), rt)
	case *aproto.Payload_Call:
		t, err = TypeOfCall(val.Call, rt)
	case *aproto.Payload_ListBool:
		t = &typing.ListType{NestedType: typing.TypeBool, Shape: val.ListBool.Shape}
	case *aproto.Payload_ListBytes:
		t = &typing.ListType{NestedType: typing.TypeBytes, Shape: val.ListBytes.Shape}
	case *aproto.Payload_ListInt:
		t = &typing.ListType{NestedType: typing.TypeInt, Shape: val.ListInt.Shape}
	case *aproto.Payload_ListUint:
		t = &typing.ListType{NestedType: typing.TypeUint, Shape: val.ListUint.Shape}
	case *aproto.Payload_ListFloat:
		t = &typing.ListType{NestedType: typing.TypeFloat, Shape: val.ListFloat.Shape}
	case *aproto.Payload_ListString:
		t = &typing.ListType{NestedType: typing.TypeString, Shape: val.ListString.Shape}
	case *aproto.Payload_ListPayload:
		t = &typing.ListType{NestedType: typing.TypeAny, Shape: val.ListPayload.Shape}
	case *aproto.Payload_Null:
		t = typing.TypeAny
	case *aproto.Payload_Singular:
		switch val.Singular.GetValue().(type) {
		case *aproto.Singular_BoolValue:
			t = typing.TypeBool
		case *aproto.Singular_BytesValue:
			t = typing.TypeBytes
		case *aproto.Singular_IntValue:
			t = typing.TypeInt
		case *aproto.Singular_UintValue:
			t = typing.TypeUint
		case *aproto.Singular_FloatValue:
			t = typing.TypeFloat
		case *aproto.Singular_StringValue:
			t = typing.TypeString
		}
	default:
		err = errors.ErrNotImplemented
	}

	return
}

func TypeOfProps(props []*aproto.Prop, rt *Runtime) (t typing.Type, err error) {
	if props == nil || rt == nil {
		return
	}

	propsType := typing.PropsType{}
	for _, prop := range props {
		propsType[prop.GetName()], err = TypeOfProp(prop, rt)
		if err != nil {
			break
		}
	}
	t = propsType
	return
}

func TypeOfProp(prop *aproto.Prop, rt *Runtime) (t typing.Type, err error) {
	if prop == nil || rt == nil {
		return
	}

	if err = rt.EntryToNode(prop); err != nil {
		return
	}
	defer func() { rt.ExitFromNode(prop, err) }()

	t, err = TypeOfPayload(prop.GetValue(), rt)
	if err != nil {
		return
	}

	if nota := prop.GetSpec().GetType(); nota != "" {
		var targetT typing.Type
		targetT, err = typing.Parse(nota)
		if err == nil {
			err = typing.Check(targetT, t)
		}
	}

	return
}

func TypeOfCall(call *aproto.Call, rt *Runtime) (t typing.Type, err error) {
	if call == nil || rt == nil {
		return
	}

	var (
		args []*aproto.Prop
	)

	if fn := rt.BuildIn().Get(call.Name); fn != nil {
		t, err = fn.Type(args, call.Params, rt)
	} else {
		err = errors.ErrFuncNotFound
	}

	return
}
