package typing

import (
	"reflect"
)

type Type interface {
	GetNotation() string
	Check(inType Type) (ok bool)
}

type BasicType struct {
	notation   string
	nativeType reflect.Kind
	protoType  reflect.Kind
}

func newBasicType(notation string, nativeType interface{}, protoType interface{}) *BasicType {
	return &BasicType{
		notation:   notation,
		nativeType: reflect.TypeOf(nativeType).Kind(),
		protoType:  reflect.TypeOf(protoType).Kind(),
	}
}

func (t *BasicType) GetNotation() string {
	return t.notation
}

func (t *BasicType) Check(inType Type) bool {
	basicInType, ok := inType.(*BasicType)
	if !ok {
		return ok
	}

	return t.nativeType == basicInType.nativeType
}

type ListType struct {
	notation   string
	NestedType Type
	protoType  reflect.Kind
	Shape      []int64
}

func (t *ListType) GetNotation() string {
	return t.notation
}

func (t *ListType) Check(inType Type) (ok bool) {
	var listInType *ListType
	if listInType, ok = inType.(*ListType); !ok {
		return
	}

	// NOTE: check shapes
	// if ok = len(t.Shape) == len(listInType.Shape); ok {
	// 	for i := range t.Shape {
	// 		if ok = t.Shape[i] == listInType.Shape[i]; !ok {
	// 			break
	// 		}
	// 	}
	// }

	if ok {
		ok = t.NestedType.Check(listInType.NestedType)
	}

	return
}

type SetType struct {
	notation string
	Types    []Type
}

func (t *SetType) GetNotation() string {
	return t.notation
}

func (t *SetType) Check(inType Type) (ok bool) {
	for _, nestedType := range t.Types {
		if ok = nestedType.Check(inType); ok {
			break
		}
	}
	return ok
}

type PropsType map[string]Type

func (t PropsType) GetNotation() string {
	return "props"
}

func (t PropsType) Check(inType Type) (ok bool) {
	var mapInType PropsType
	if mapInType, ok = inType.(PropsType); !ok {
		return
	}
	for key, valType := range t {
		if inType, ok = mapInType[key]; ok {
			ok = valType.Check(inType)
		}
		if !ok {
			break
		}
	}

	return
}

type CustomType struct {
	FnNotation func() string
	FnValidate func(inType Type) (ok bool)
}

func (t *CustomType) GetNotation() string {
	return t.FnNotation()
}

func (t *CustomType) Check(inType Type) (ok bool) {
	return t.FnValidate(inType)
}
