package store

import (
	"github.com/cosmos/cosmos-sdk/store/cachekv"
	"github.com/cosmos/cosmos-sdk/store/prefix"
	st_types "github.com/cosmos/cosmos-sdk/store/types"

	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/errors"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang/proto"
)

type objectStore struct {
	store st_types.KVStore
	reg   *Registry
}

func (st *objectStore) get(id *proto.ID, cached bool) (obj *Object, err error) {
	var objStore st_types.KVStore = prefix.NewStore(st.store, id.Bytes())
	if cached {
		objStore = cachekv.NewStore(objStore)
	}
	obj = makeObject(id, objStore)
	if obj.IsEmpty() {
		err = errors.ErrObjectNotFound
	}
	return
}

func (st *objectStore) New() (obj *Object) {
	if st == nil {
		return
	}

	if st.reg != nil {
		st.reg.Lock()
		defer st.reg.Unlock()
	}

	for obj == nil || !obj.IsEmpty() {
		id := proto.GenerateID()
		ok := true
		if st.reg != nil {
			_, ok = st.reg.lockedObjects[id.Base64()]
			ok = !ok
		}
		if ok {
			obj = makeObject(id, prefix.NewStore(st.store, id.Bytes()))
		}
	}
	if st.reg != nil {
		st.reg.lockedObjects[obj.ID().Base64()] = -1
	}

	obj.Balance = NewBalance(nil)
	obj.WriteBalance()
	obj.Set(&aproto.Prop{
		Name:  "id",
		Value: proto.BytesPayload(obj.id.Bytes()),
	})

	return
}

func (st *objectStore) Get(id *proto.ID) (obj *Object, err error) {
	if st == nil {
		return
	}

	if st.reg == nil {
		obj, err = st.get(id, false)
	} else if err = st.reg.lock(id); err == nil {
		obj, err = st.get(id, false)
	}

	return
}

func (st *objectStore) GetForRead(id *proto.ID) (obj *Object, err error) {
	if st == nil || st.reg == nil {
		return
	}

	if st.reg == nil {
		obj, err = st.get(id, true)
	} else if err = st.reg.check(id); err == nil {
		obj, err = st.get(id, true)
	}

	return
}

type Object struct {
	st_types.KVStore
	id *proto.ID

	*Balance
}

func makeObject(id *proto.ID, st st_types.KVStore) (object *Object) {
	object = &Object{
		KVStore: st,
		id:      id,
	}
	return
}

func (obj *Object) ID() *proto.ID {
	if obj == nil {
		return nil
	}
	return obj.id
}

func (obj *Object) IsEmpty() bool {
	if obj == nil {
		return true
	}
	return !obj.Has("id")
}

func (obj *Object) Has(key string) bool {
	if obj == nil {
		return false
	}
	return obj.KVStore.Has([]byte(key))
}

func (obj *Object) Get(key string) (prop *aproto.Prop, err error) {
	if obj == nil {
		return
	}

	bts := obj.KVStore.Get([]byte(key))
	if len(bts) == 0 {
		err = errors.ErrPropNotFound
		return
	}

	prop = &aproto.Prop{}
	err = proto.Unmarshal(bts, prop)

	return
}

func (obj *Object) Set(prop *aproto.Prop) error {
	if obj == nil {
		return nil
	}

	bts, err := proto.Marshal(prop)
	if err != nil {
		return err
	}

	obj.KVStore.Set([]byte(prop.Name), bts)

	return nil
}

func (obj *Object) Iterate(fn func(prop *aproto.Prop) bool) (err error) {
	if obj == nil || fn == nil {
		return
	}
	iter := obj.Iterator(nil, nil)

	prop := &aproto.Prop{}
	for iter.Valid() {
		err = proto.Unmarshal(iter.Value(), prop)
		if err != nil {
			break
		}

		if !fn(prop) {
			break
		}

		iter.Next()
	}

	return
}

func (obj *Object) FromProps(props []*aproto.Prop) (err error) {
	if obj == nil {
		return
	}

	for _, prop := range props {
		err = obj.Set(prop)
		if err != nil {
			break
		}
	}
	return
}

func (obj *Object) LoadBalance() error {
	if obj == nil {
		return nil
	}

	if obj.Balance == nil {
		balance, err := obj.Get("balance")
		if err != nil {
			return err
		}
		funds, err := proto.BigUintFromBytes(balance.Value.GetSingular().GetBytesValue())
		if err != nil {
			return err
		}
		obj.Balance = NewBalance(funds)
	}
	return nil
}

func (obj *Object) WriteBalance() error {
	if obj == nil || obj.Balance == nil {
		return nil
	}

	return obj.Set(&aproto.Prop{
		Name:  "balance",
		Value: obj.Balance.free.Payload(),
	})
}
