package store

import (
	"github.com/cosmos/cosmos-sdk/store/cachekv"
	"github.com/cosmos/cosmos-sdk/store/rootmulti"
	st_types "github.com/cosmos/cosmos-sdk/store/types"
	dbm "github.com/tendermint/tm-db"
)

var (
	objectStoreKey = st_types.NewKVStoreKey("object")
	indexStoreKey  = st_types.NewKVStoreKey("index")
	rollupStoreKey = st_types.NewKVStoreKey("rollup")
)

type Store struct {
	iidap_id string
	db       dbm.DB

	store            st_types.CommitMultiStore
	substoresMounted bool
	objectStore      *objectStore
	indexStore       *indexStore
	rollupStore      st_types.KVStore
}

func MakeStore(iidap_id string, db dbm.DB) (st *Store, err error) {
	if db == nil {
		db = dbm.NewMemDB()
	}
	db = dbm.NewPrefixDB(db, []byte(iidap_id))

	st = &Store{
		iidap_id: iidap_id,
		db:       db,
	}

	err = st.init()

	return
}

func (st *Store) init() (err error) {
	if st.store == nil {
		st.store = rootmulti.NewStore(st.db)

		st.store.MountStoreWithDB(objectStoreKey, st_types.StoreTypeIAVL, nil)
		st.store.MountStoreWithDB(indexStoreKey, st_types.StoreTypeIAVL, nil)
		st.store.MountStoreWithDB(rollupStoreKey, st_types.StoreTypeIAVL, nil)

		err = st.store.LoadLatestVersion()
		if err != nil {
			return
		}
	}

	if !st.substoresMounted {
		st.objectStore = &objectStore{
			store: st.store.GetCommitKVStore(objectStoreKey),
		}
		st.indexStore = &indexStore{st.store.GetCommitKVStore(indexStoreKey)}
		st.rollupStore = st.store.GetKVStore(rollupStoreKey)
		st.substoresMounted = true
	}
	return
}

func (st *Store) ExecWrapper(reg *Registry) (wrapped *Store, err error) {
	wrapped = &Store{
		iidap_id: st.iidap_id,
		store:    st.store,
		objectStore: &objectStore{
			store: cachekv.NewStore(st.objectStore.store),
			reg:   reg,
		},
		indexStore: &indexStore{cachekv.NewStore(st.indexStore.store)},
	}

	return
}

func (st *Store) LastCommitID() st_types.CommitID {
	return st.store.LastCommitID()
}

func (st *Store) Commit() st_types.CommitID {
	return st.store.Commit()
}

func (st *Store) Write() {
	if cacheSt, ok := st.objectStore.store.(st_types.CacheKVStore); ok {
		cacheSt.Write()
	}
	if cacheSt, ok := st.indexStore.store.(st_types.CacheKVStore); ok {
		cacheSt.Write()
	}
}

func (st *Store) Object() *objectStore {
	return st.objectStore
}

func (st *Store) Index() *indexStore {
	return st.indexStore
}

func (st *Store) Rollup() st_types.KVStore {
	return st.rollupStore
}
