package subscribe

import (
	"sync"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang/proto"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/utils"
	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"
	"go.uber.org/zap"
)

var logger = utils.GetLogger("SUBSCRIBE_MANAGER")

// NOTE: For better performance on a high-core (32+) machine, sync.Map can be used
type subscibersMap map[string]*Subscriber

type Manager struct {
	subscribers subscibersMap
	mutex       sync.RWMutex
}

var defaultManager *Manager = nil

func DefaultManager() *Manager {
	if defaultManager == nil {
		defaultManager = &Manager{
			subscribers: make(subscibersMap),
			mutex:       sync.RWMutex{},
		}
	}
	return defaultManager
}

// MANAGE SUBSCRIBERS
// ==================

func (m *Manager) GetClientSubscriber(clientID *proto.ID) *Subscriber {
	m.mutex.RLock()
	defer m.mutex.RUnlock()

	return m.subscribers[clientID.Base64()]
}

func (m *Manager) RegisterSubscriber(sbr *Subscriber) {
	m.mutex.Lock()
	defer m.mutex.Unlock()

	if oldSbr, ok := m.subscribers[sbr.clientID.Base64()]; ok {
		oldSbr.close()
	}

	m.subscribers[sbr.clientID.Base64()] = sbr
	logger.Debug("Subscriber registered", zap.String("client_id", sbr.clientID.Base64()))
}

func (m *Manager) UnregisterSubscriber(sbr *Subscriber) {
	m.mutex.Lock()
	defer m.mutex.Unlock()

	if curSbr, ok := m.subscribers[sbr.clientID.Base64()]; ok && curSbr == sbr {
		sbr.close()
		delete(m.subscribers, sbr.clientID.Base64())
		logger.Debug("Subscriber unregistered", zap.String("client_id", sbr.clientID.Base64()))
	}
}

func (m *Manager) CloseClientSubscriber(clientID *proto.ID) {
	m.mutex.Lock()
	defer m.mutex.Unlock()

	if sbr, ok := m.subscribers[clientID.Base64()]; ok {
		sbr.close()
		delete(m.subscribers, clientID.Base64())
		logger.Debug("Subscriber unregistered and closed", zap.String("client_id", sbr.clientID.Base64()))
	}
}

func (m *Manager) CloseAllSubscribers() {
	m.mutex.Lock()
	defer m.mutex.Unlock()

	for clientID, sbr := range m.subscribers {
		sbr.close()
		delete(m.subscribers, clientID)
	}
	logger.Debug("All subscribers unregistered and closed")
}

func (m *Manager) SubscribersClientIDs() []string {
	m.mutex.RLock()
	defer m.mutex.RUnlock()

	clientIDs := make([]string, 0, len(m.subscribers))

	for clientID := range m.subscribers {
		clientIDs = append(clientIDs, clientID)
	}
	logger.Debug("Active subscribers", zap.Strings("client_keys", clientIDs))
	return clientIDs
}

// MANAGE MESSAGES
// =============

func (m *Manager) BroadcastMessage(msg *aproto.Msg) {
	clientID := proto.NewID(msg.Target)
	if sbr := m.GetClientSubscriber(clientID); sbr != nil {
		if err := sbr.sendMsg(msg); err != nil {
			m.UnregisterSubscriber(sbr)
			logger.Error("Fail to SendMessages",
				zap.String("client_key", clientID.Base64()),
				zap.Error(err),
			)
		}
	}
}

func (m *Manager) BroadcastMessages(msgs []*aproto.Msg) {
	for _, msg := range msgs {
		m.BroadcastMessage(msg)
	}
}
