package launch

import (
	"context"
	"os"
	"os/signal"
	"syscall"

	"go.uber.org/zap"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/api/grpc"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/api/ws"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/dapp"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/exec"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/rollup"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/store"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/subscribe"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/utils"
)

var (
	logger = utils.GetLogger("LAUNCH_ENGINE")
)

type LaunchEngine struct {
	errChan chan error
	sigChan chan os.Signal

	ctx       context.Context
	ctxCancel context.CancelFunc

	execController *exec.ExecController
	subscribeMngr  *subscribe.Manager

	grpcServer *grpc.Server
	wsServer   *ws.Server
}

func Make() (engn *LaunchEngine, err error) {

	st, err := store.MakeStore(dapp.DApp.ID().Base64(), nil)
	if err != nil {
		return
	}

	var (
		ctx, cancelFn  = context.WithCancel(context.Background())
		rollupProc     = rollup.NewRollupProcessor(st)
		subscribeMngr  = subscribe.DefaultManager()
		execController = exec.NewExecController(subscribeMngr, st, rollupProc)
	)

	engn = &LaunchEngine{
		errChan:        make(chan error),
		sigChan:        make(chan os.Signal, 1),
		ctx:            ctx,
		ctxCancel:      cancelFn,
		execController: execController,
		subscribeMngr:  subscribeMngr,
		grpcServer:     grpc.NewServer(ctx),
		wsServer:       ws.NewServer(ctx),
	}
	return
}

func (engn *LaunchEngine) Start() error {
	err := engn.startModules()
	if err != nil {
		return err
	}
	go engn.handleSignal()
	return engn.wait()
}

func (engn *LaunchEngine) Stop() {
	logger.Info("Called Engine.Stop")
	engn.sigChan <- os.Interrupt
}

func (engn *LaunchEngine) startModules() (err error) {
	logger.Info("Start ExecController...")
	go func() {
		engn.errChan <- engn.execController.Start()
	}()

	logger.Info("Start GrpcServer...")
	go func() {
		engn.errChan <- engn.grpcServer.Start(engn.subscribeMngr, engn.execController)
	}()

	logger.Info("Start WSServer...")
	go func() {
		engn.errChan <- engn.wsServer.Start(engn.subscribeMngr, engn.execController)
	}()

	return
}

func (engn *LaunchEngine) handleSignal() {
	signal.Notify(engn.sigChan,
		syscall.SIGHUP,
		syscall.SIGINT,
		syscall.SIGTERM,
		syscall.SIGQUIT)

	sig, ok := <-engn.sigChan
	if ok {
		logger.Info("Received signal", zap.String("signal", sig.String()))
	} else {
		logger.Info("Signal channel closed")
	}

	logger.Info("Stop engine...")

	engn.grpcServer.Stop()
	engn.wsServer.Stop()
	engn.execController.Stop()

	os.Exit(0)
}

func (engn *LaunchEngine) wait() (err error) {
	var ok bool
	for {
		err, ok = <-engn.errChan
		if !ok {
			logger.Error("Error channel closed")
			engn.Stop()
			break
		} else if err != nil {
			logger.Error("Error handled", zap.Error(err))
			engn.Stop()
			break
		}
	}
	return
}
