package main

import (
	"fmt"
	"log"
	"sync"

	"github.com/gdamore/tcell/v2"
	"github.com/pgavlin/femto"
	"github.com/rivo/tview"
	googleGRPC "google.golang.org/grpc"

	aproto "gitlab.domi.do/avvy/backend-protocol/proto/avvyprotocol"

	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/api/grpc"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang/compiler"
	"gitlab.domi.do/avvy/avvy-l3/l3_node_golang/core/lang/proto"
)

type shell struct {
	addr      string
	clientKey string

	targetObjectID string
	handlerName    string

	app        *tview.Application
	buf        *femto.Buffer
	editorView *femto.View
	out        chan string
	outView    *tview.TextView

	mutex sync.RWMutex

	conn   *grpc.Client
	stream aproto.L3Service_StreamClient
}

func (sh *shell) print(text string) {
	sh.out <- text
}

func (sh *shell) err(err error) {
	sh.out <- fmt.Sprintf("[error]: %s", err.Error())
}

func (sh *shell) init() {
	sh.app = tview.NewApplication().EnableMouse(true)

	connView := tview.NewFlex().SetDirection(tview.FlexColumn)
	{
		connView.AddItem(
			tview.NewTextView().SetText("Node address").SetTextAlign(1),
			0, 1, false,
		)
		connView.AddItem(
			tview.NewInputField().SetText(sh.addr).SetChangedFunc(func(text string) {
				sh.addr = text
			}),
			0, 1, false,
		)
		connView.AddItem(
			tview.NewTextView().SetText("Client key").SetTextAlign(1),
			0, 1, false,
		)
		connView.AddItem(
			tview.NewInputField().SetText(sh.clientKey).SetChangedFunc(func(text string) {
				sh.clientKey = text
			}),
			0, 1, false,
		)

		button := tview.NewButton("Connect").SetSelectedFunc(func() {
			sh.mutex.Lock()
			defer sh.mutex.Unlock()

			if sh.conn != nil {
				sh.print("Drop existed connection...")
				sh.conn.Close()
			}

			sh.print("Connect...")
			var err error
			sh.conn, err = grpc.NewClient(sh.addr, googleGRPC.WithInsecure())
			if err != nil {
				sh.err(err)
				return
			}

			clientID, err := proto.MakeIDFromBase64(sh.clientKey)
			if err != nil {
				sh.err(err)
				return
			}

			sh.stream, err = sh.conn.StartStream(clientID, nil)
			if err != nil {
				sh.err(err)
				return
			}
			sh.print("Connected")

			go func() {
				for {
					txn, err := sh.stream.Recv()
					if err != nil {
						break
					}
					sh.print(txn.String())
				}
			}()
		})
		button.SetLabelColor(tcell.ColorWhite)
		button.SetBackgroundColor(tcell.ColorOrange)
		connView.AddItem(button, 0, 1, false)
	}

	msgSendView := tview.NewFlex().SetDirection(tview.FlexColumn)
	{
		msgSendView.AddItem(
			tview.NewTextView().SetText("Target ObjectID").SetTextAlign(1),
			0, 1, false,
		)
		msgSendView.AddItem(
			tview.NewInputField().SetText(sh.targetObjectID).SetChangedFunc(func(text string) {
				sh.targetObjectID = text
			}),
			0, 1, false,
		)
		msgSendView.AddItem(
			tview.NewTextView().SetText("Handler").SetTextAlign(1),
			0, 1, false,
		)
		msgSendView.AddItem(
			tview.NewInputField().SetText(sh.handlerName).SetChangedFunc(func(text string) {
				sh.handlerName = text
			}),
			0, 1, false,
		)

		button := tview.NewButton("Send").SetSelectedFunc(func() {

			sh.print("Compile message...")
			props, err := compiler.Compile(sh.buf.String())
			if err != nil {
				sh.err(err)
				return
			}
			sh.print("Compiled")

			clientID, err := proto.MakeIDFromBase64(sh.clientKey)
			if err != nil {
				sh.err(err)
				return
			}

			targetID, err := proto.MakeIDFromBase64(sh.clientKey)
			if err != nil {
				sh.err(err)
				return
			}

			msg := &aproto.Msg{
				Type:   sh.handlerName,
				Sender: clientID.Bytes(),
				Target: targetID.Bytes(),
				Props:  props.Props(),
			}

			txn := &aproto.TxnRequest{
				AuthInfo: &aproto.AuthInfo{
					Signers: []*aproto.Signer{
						{
							PublicKey: clientID.Bytes(),
						},
					},
				},
				Messages: []*aproto.Msg{msg},
			}

			sh.print("Send message...")
			err = sh.stream.Send(txn)
			if err != nil {
				sh.err(err)
			}
			sh.print("Sended")
		})
		button.SetLabelColor(tcell.ColorWhite)
		button.SetBackgroundColor(tcell.ColorOrange)
		msgSendView.AddItem(button, 0, 1, false)
	}

	sh.outView = tview.NewTextView()
	go func() {
		for {
			out, ok := <-sh.out
			if !ok {
				break
			}
			sh.outView.Write([]byte(fmt.Sprintf("%s\n", out)))
		}
	}()

	sh.buf = femto.NewBufferFromString("", "")
	sh.editorView = femto.NewView(sh.buf)

	{
		root := tview.NewFlex().SetDirection(tview.FlexRow)

		root.AddItem(
			tview.NewFlex().SetDirection(tview.FlexColumn).
				AddItem(sh.editorView, 0, 1, false).
				AddItem(sh.outView, 0, 1, false),
			0, 1, false,
		)

		root.AddItem(
			tview.NewFlex().SetDirection(tview.FlexColumn).
				AddItem(msgSendView, 0, 1, false).
				AddItem(connView, 0, 1, false),
			1, 0, false,
		)

		sh.app.SetRoot(root, true)
	}

}

func (sh *shell) run() error {
	return sh.app.Run()
}

func main() {
	sh := &shell{
		addr:           "127.0.0.1:80",
		clientKey:      "0000",
		targetObjectID: "0000",
		handlerName:    "help",
		out:            make(chan string, 10),
		mutex:          sync.RWMutex{},
	}
	sh.init()
	if err := sh.run(); err != nil {
		log.Fatalln(err)
	}
}
